from .._ast import RoleEnd, RoleStart
from ._block import Block, block


# TODO HN: Add a docstring to better describe arbitrary role functions
def role(role: str) -> Block:
    return block(
        name=None,
        opener=RoleStart(role),
        closer=RoleEnd(role),
    )


def system() -> Block:
    """Indicate the 'system' prompt

    A convention has grown up around 'chat' APIs that
    prompts are split into three parts: system, user
    and assistant.
    This indicates the start of a 'system' block, which
    provides background information to the LLM.

        >>> with system():
        >>>     lm += "A system prompt"

    """
    return role("system")


def user() -> Block:
    """Indicate the 'user' prompt

    A convention has grown up around 'chat' APIs that
    prompts are split into three parts: system, user
    and assistant.
    This indicates the start of a 'user' block, which
    provides input to the LLM from the user.

        >>> with user():
        >>>     lm += "What the user said"

    """
    return role("user")


def assistant() -> Block:
    """Indicate the 'assistant' prompt

    A convention has grown up around 'chat' APIs that
    prompts are split into three parts: system, user
    and assistant.
    This indicates the start of an 'assistant' block, which
    marks LLM response (or where the LLM will generate
    the next response).

        >>> with assistant():
        >>>     lm += gen(name="model_output", max_tokens=20)

    """
    return role("assistant")



# import functions that can be called directly
from ._gen import gen, call_tool, will_gen, regex
from ._image import image, gen_image
from ._audio import audio, gen_audio
from ._video import video, gen_video
from ._capture import capture

# core grammar functions
from .._grammar import select
from .._grammar import with_temperature
from .._grammar import string
from .._grammar import token_limit

# context blocks
from ._block import block
from ._role import role, system, assistant, user #, function, instruction, indent_roles
from ._format import monospace
from ._silent import silent
from ._set_var import set_var
from ._set_attribute import set_attribute

# from ..models._model import context_free

# stateless library functions
from ._sequences import one_or_more, zero_or_more, at_most_n_repeats, exactly_n_repeats, sequence
from ._prefix_tree import prefix_tree
from ._substring import substring
from ._optional import optional
from ._tool import Tool
from ._json import json
from ._ebnf import lark, gbnf_to_lark



""" Visualization related to trace. """

import base64
import json
from typing import Optional, Dict
from .._schema import GenToken

from ..visual._message import TokensMessage
from ..trace import (
    TextOutput,
    TraceNode,
    TraceHandler,
    RoleOpenerInput,
    RoleCloserInput,
    ImageOutput,
)
import html


def trace_node_to_html(
    node: TraceNode, prettify_roles=False, complete_msg: TokensMessage = None
) -> str:
    """Represents trace path as html string.

    Args:
        node: Trace node that designates the end of a trace path for HTML output.
        prettify_roles: Enables prettier formatting for roles.
        complete_msg: Output message received on completion of engine.

    Returns:
        HTML string of trace path as html.
    """
    buffer = []
    node_path = list(node.path())
    active_role: Optional[TraceNode] = None

    prob_idx = 0
    # remaining_text = ""
    full_text = ""
    if complete_msg:
        for token in complete_msg.tokens:
            full_text += token.text

    for i, node in enumerate(node_path):
        if isinstance(node.input, RoleOpenerInput):
            active_role = node
        elif isinstance(node.input, RoleCloserInput):
            active_role = node

        if isinstance(node.output, TextOutput):
            if active_role is not None:
                if isinstance(active_role.input, RoleOpenerInput) and prettify_roles:
                    role_name = active_role.input.name
                    fmt = f"<div style='display: flex; border-bottom: 1px solid rgba(127, 127, 127, 0.2);  justify-content: center; align-items: center;'><div style='flex: 0 0 80px; opacity: 0.5;'>{role_name.lower()}</div><div style='flex-grow: 1; padding: 5px; padding-top: 10px; padding-bottom: 10px; margin-top: 0px; white-space: pre-wrap; margin-bottom: 0px;'>"
                    buffer.append(fmt)
                if not prettify_roles:
                    buffer.append("<span style='background-color: rgba(255, 180, 0, 0.3); border-radius: 3px;'>")

            if not (active_role and prettify_roles):
                attr = node.output

                fmt = ""
                if not complete_msg:
                    if attr.is_generated:
                        # fmt = f"<span style='background-color: rgba({165 * (1 - attr.prob)}, {165 * attr.prob}, 0, 0.15); border-radius: 3ps;' title='{attr.prob}'>{html.escape(attr.value)}</span>"
                        fmt = f"<span style='background-color: rgba({0}, {255}, {0}, 0.15); border-radius: 3ps;' title='{attr.prob}'>{html.escape(attr.value)}</span>"
                    elif attr.is_force_forwarded:
                        fmt = f"<span style='background-color: rgba({0}, {0}, {255}, 0.15); border-radius: 3ps;' title='{attr.prob}'>{html.escape(attr.value)}</span>"
                    else:
                        # fmt = f"{html.escape(attr.value)}"
                        fmt += f"<span style='background-color: rgba({255}, {255}, {255}, 0.15); border-radius: 3ps;' title='{attr.prob}'>{html.escape(attr.value)}</span>"
                else:
                    # switch to token-based
                    # cell_tokens = attr.tokens
                    # for token in cell_tokens:
                    #     # assert token.token == complete_msg.tokens[prob_idx].token, f"Token mismatch {token.token} != {complete_msg.tokens[prob_idx].token}"
                    #     if token.token_id != complete_msg.tokens[prob_idx].token_id:
                    #         if remaining_text + token.text != complete_msg.tokens[prob_idx].text:
                    #             remaining_text += token.text
                    #             continue
                    #         else:
                    #             remaining_text = ""

                    #     token_str = complete_msg.tokens[prob_idx].text
                    #     prob = complete_msg.tokens[prob_idx].prob
                    #     top_k = {}
                    #     # find the correct token
                    #     for _item in complete_msg.tokens[prob_idx].top_k:
                    #         top_k[f"{_item.text}"] = f"{_item.prob} - Masked: {_item.is_masked}"
                    #     top_k = json.dumps(top_k, indent=2)

                    #     latency = f"{complete_msg.tokens[prob_idx].latency_ms:.2f}"

                    #     if complete_msg.tokens[prob_idx].is_generated:
                    #         fmt += f"<span style='background-color: rgba({0}, {127 + int(127 * prob)}, {0}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k}\nlatency_ms: {latency}'>{html.escape(token_str)}</span>"
                    #     elif complete_msg.tokens[prob_idx].is_force_forwarded:
                    #         fmt += f"<span style='background-color: rgba({0}, {0}, {127 + int(127 * prob)}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k}\nlatency_ms: {latency}'>{html.escape(token_str)}</span>"
                    #     else:
                    #         fmt += f"<span style='background-color: rgba({255}, {255}, {255}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k}'>{html.escape(token_str)}</span>"

                    #     full_text = full_text[len(token_str) :]
                    #     prob_idx += 1

                    chunk_text = attr.value
                    # find tokens in complete message that cover this chunk
                    tokens: list[GenToken] = []
                    _idx = prob_idx
                    tokens_text = ""
                    while _idx < len(complete_msg.tokens):
                        _token = complete_msg.tokens[_idx]
                        tokens_text += _token.text
                        tokens.append(_token)
                        if chunk_text in tokens_text:
                            break
                        _idx += 1

                    assert (
                        chunk_text in tokens_text
                    ), f"Token mismatch {tokens_text} != {chunk_text}"

                    start_idx = tokens_text.index(chunk_text)
                    remaining_text = tokens_text[start_idx + len(chunk_text) :]

                    if remaining_text:
                        # remove the last tokens
                        tokens.pop(-1)

                    # update prob_idx
                    prob_idx += len(tokens)

                    for token in tokens:
                        token_str = token.text
                        prob = token.prob
                        top_k = {}
                        # find the correct token
                        for _item in token.top_k:
                            top_k[f"{_item.text}"] = f"{_item.prob} - Masked: {_item.is_masked}"
                        top_k = json.dumps(top_k, indent=2)

                        latency = f"{token.latency_ms:.2f}"

                        if token.is_generated:
                            fmt += f"<span style='background-color: rgba({0}, {127 + int(127 * prob)}, {0}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k}\nlatency_ms: {latency}'>{html.escape(token_str)}</span>"
                        elif token.is_force_forwarded:
                            fmt += f"<span style='background-color: rgba({0}, {0}, {127 + int(127 * prob)}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k}\nlatency_ms: {latency}'>{html.escape(token_str)}</span>"
                        else:
                            fmt += f"<span style='background-color: rgba({255}, {255}, {255}, 0.15); border-radius: 3ps;' title='Token: \"{token_str}\" : {prob}\nTop-k: {top_k}'>{html.escape(token_str)}</span>"

                buffer.append(fmt)

            if active_role is not None:
                if not prettify_roles:
                    buffer.append("</span>")
                if isinstance(active_role.input, RoleCloserInput) and prettify_roles:
                    buffer.append(f"</div></div>")
                active_role = None
        elif isinstance(node.output, ImageOutput):
            encoded = base64.b64encode(node.output.value).decode()
            buffer.append(
                f'<img src="data:image/png;base64,{encoded}" style="max-width" 400px; vertical-align: middle; margin: 4px;">'
            )

    buffer.insert(
        0,
        "<pre style='margin: 0px; padding: 0px; vertical-align: middle; padding-left: 8px; margin-left: -8px; border-radius: 0px; border-left: 1px solid rgba(127, 127, 127, 0.2); white-space: pre-wrap; font-family: ColfaxAI, Arial; font-size: 15px; line-height: 23px;'>",
    )
    buffer.append("</pre>")
    return "".join(buffer)


def trace_node_to_str(node: TraceNode) -> str:
    """Visualize output attributes of a trace node up to the root.

    Users should not be accessing this function. For debugging purposes.

    Args:
        node: The trace node to visualize.
    Returns:
        Output as string.
    """
    buffer = []
    for node in node.path():
        if isinstance(node.output, TextOutput):
            buffer.append(str(node.output))
    return "".join(buffer)


def display_trace_tree(trace_handler: TraceHandler) -> None:
    """Visualize tree of a trace node going down to all its leaves.

    Users should not normally be accessing this function. For debugging purposes.

    Args:
        trace_handler: Trace handler needed to pull user-defined identifiers of trace nodes.
    """
    from anytree import Node, RenderTree

    root = trace_handler.root()
    trace_viz_map: Dict[TraceNode, Node] = {}
    for node in root.traverse(bfs=False):
        viz_parent = trace_viz_map.get(node.parent, None)
        viz_node = Node(f"{trace_handler.node_id_map[node]}:{node!r}", parent=viz_parent)
        trace_viz_map[node] = viz_node
    viz_root = trace_viz_map[root]

    for pre, fill, node in RenderTree(viz_root):
        tree_str = "%s%s" % (pre, node.name)
        print(tree_str)



""" Renderer is responsible for displaying output.

Our main focus is on jupyter notebooks and later terminal.
"""
# TODO(nopdive): Implementation for terminals & append-only text displays.
# NOTE(nopdive): Testing this notebook related components is tricky. Should figure this out at some point.

import asyncio
import logging
import weakref
from typing import Optional, Tuple
from asyncio import Queue
from functools import partial
import traceback

from . import MetricMessage
from ._environment import Environment
from ._jupyter import ipy_handle_event_once
from ._message import ExecutionCompletedMessage, \
    deserialize_message, serialize_message, ClientReadyAckMessage, ExecutionStartedMessage, OutputRequestMessage
from .._utils import log_cleanup
from ..trace import TraceHandler
from ..visual import GuidanceMessage, TraceMessage, ResetDisplayMessage, ClientReadyMessage
from warnings import warn

try:
    from IPython.display import clear_output, display, HTML
    from IPython import get_ipython

    ipython_imported = True
except ImportError:
    ipython_imported = False


try:
    import stitch

    stitch_installed = True
except ImportError:
    stitch_installed = False

logger = logging.getLogger(__name__)

DEFAULT_TOPIC = "default"


class Renderer:
    """Renders guidance model to a visual medium."""

    def __init__(self):
        """Initializes. """

    def update(self, message: GuidanceMessage, topic: str = DEFAULT_TOPIC) -> None:
        """ Updates renderer with incoming message.

        Args:
            message: Incoming message.
            topic: Topic to update.
        """
        raise NotImplementedError("Update not implemented.")


def _create_stitch_widget():
    from stitch import StitchWidget
    import importlib.resources as resources
    import guidance

    if _create_stitch_widget.src_doc_template is None:
        path = resources.files(guidance) / 'resources' / 'graphpaper-inline.html'
        with path.open("r") as f:
            _create_stitch_widget.src_doc_template = f.read()
    w = StitchWidget()
    w.initial_width = "100%"
    w.initial_height = "auto"
    w.srcdoc = _create_stitch_widget.src_doc_template
    weakref.finalize(w, log_cleanup, f"stitch({id(w)})")

    return w


_create_stitch_widget.src_doc_template = None


def _cleanup(recv_queue: Optional[Queue], send_queue: Optional[Queue], log_msg: str, exchange_cb) -> None:
    from ..registry import get_bg_async, get_exchange

    log_cleanup(log_msg)
    if recv_queue is not None:
        get_bg_async().call_soon_threadsafe(send_queue.put_nowait, None)
    if send_queue is not None:
        get_bg_async().call_soon_threadsafe(recv_queue.put_nowait, None)

    get_exchange().unsubscribe(exchange_cb)


async def _create_queue() -> Queue:
    # This will run in the visual thread.
    return Queue()


def _on_stitch_clientmsg(recv_queue_weakref: weakref.ReferenceType["Queue"], change: dict) -> None:
    from ..registry import get_bg_async

    # NOTE(nopdive): Widget callbacks do not print to stdout/stderr nor module log.
    recv_queue = recv_queue_weakref()
    if recv_queue is not None:
        get_bg_async().call_soon_threadsafe(recv_queue.put_nowait, change['new'])


def _on_cell_completion(renderer_weakref: weakref.ReferenceType["JupyterWidgetRenderer"], info) -> None:
    logger.debug(f"CELL:executed")
    try:
        renderer = renderer_weakref()
        if renderer is None:
            return

        message = ExecutionCompletedMessage(
            last_trace_id=renderer.last_trace_id,
            is_err=info.error_in_exec is not None,
        )
        renderer.update(message)
    except Exception as _:
        logger.error(f"CELL_COMPLETE:{traceback.format_exc()}")


async def _handle_recv_messages(renderer_weakref: weakref.ReferenceType["JupyterWidgetRenderer"], queue_weakref: weakref.ReferenceType["Queue"]) -> None:
    logger.debug("RECV:init")
    from ..registry import get_exchange
    from ..registry import get_bg_async

    while True:
        try:
            queue = queue_weakref()
            if queue is None:
                break
            value = await queue.get()

            # logger.debug(f"RECV:raw:{value}")

            if value is None:
                logger.debug("RECV:closing")
                break

            message = deserialize_message(value)
            # logger.debug(f"RECV:msg:{message}")

            renderer = renderer_weakref()
            if renderer is None:
                logger.debug("RECV:renderer early clean")
                break

            if isinstance(message, ClientReadyMessage):
                logger.debug("RECV:clientready")
                get_bg_async().call_soon_threadsafe(renderer.send_queue.put_nowait, ClientReadyAckMessage())
            elif isinstance(message, OutputRequestMessage):
                logger.debug("RECV:outputrequest")

            get_exchange().publish(message, topic=f"{DEFAULT_TOPIC}")
            renderer.recv_queue.task_done()
        except Exception as _:
            logger.error(f"RECV:err:{traceback.format_exc()}")


async def _handle_send_messages(renderer_weakref: weakref.ReferenceType["JupyterWidgetRenderer"], queue_weakref: weakref.ReferenceType["Queue"]) -> None:
    logger.debug("SEND:init")
    # NOTE(nopdive): Waiting on client cb does not work, client messages received on cell completion.
    #                Currently, we do a replay of messages on completion for client if client
    #                first receives non-zero message identifier.

    # What if we only used 1% of our brain?
    await asyncio.sleep(200 / 1000.)
    logger.debug("SEND:ready")

    while True:
        try:
            queue = queue_weakref()
            if queue is None:
                break
            message = await queue.get()
            # logger.debug(f"SEND:msg:{message}")

            if message is None:
                logger.debug("SEND:closing")
                break

            message_json = serialize_message(message)
            # logger.debug(f"SEND:json:{message_json}")

            renderer = renderer_weakref()
            if renderer is None:
                break
            if isinstance(renderer, JupyterWidgetRenderer) and renderer.stitch_widget is not None:
                # NOTE(nopdive): This at random times, appears to fire two changes instead of one change event.
                renderer.stitch_widget.kernelmsg = message_json
            else:
                logger.debug(f"SEND:jupyter:send but no widget")
            renderer.send_queue.task_done()
        except Exception as _:
            logger.error(f"SEND:err:{traceback.format_exc()}")


class JupyterWidgetRenderer(Renderer):
    """Jupyter widget renderer that is implemented via stitch package."""

    def __init__(self, trace_handler: TraceHandler) -> None:
        """ Initializes.

        Args:
            trace_handler: Trace handler of an engine.
        """
        from ..registry import get_bg_async, get_exchange

        super().__init__()

        self.stitch_widget = None
        self.last_trace_id = None

        self._trace_handler = trace_handler
        self._messages: list[GuidanceMessage] = []
        self._completed = False
        self._running = False
        self._new_widget_needed = False
        self.stitch_widget_observed = False
        self._stitch_on_clientmsg = None
        self.last_cell_session_id = None

        # Create queue and wait for instantiation
        self.send_queue: Queue = get_bg_async().run_async_coroutine(_create_queue()).result()
        self.recv_queue: Queue = get_bg_async().run_async_coroutine(_create_queue()).result()

        # Start send/recv message loops
        recv_coroutine = _handle_recv_messages(weakref.ref(self), weakref.ref(self.recv_queue))
        send_coroutine = _handle_send_messages(weakref.ref(self), weakref.ref(self.send_queue))
        self._recv_task = get_bg_async().run_async_coroutine(get_bg_async().async_task(recv_coroutine)).result()
        self._send_task = get_bg_async().run_async_coroutine(get_bg_async().async_task(send_coroutine)).result()

        # Start recv on exchange
        get_exchange().subscribe(self._on_exchange)
        weakref.finalize(self, _cleanup, self.recv_queue, self.send_queue, f"renderer({id(self)})", self._on_exchange)


    def _on_exchange(self, message: GuidanceMessage) -> None:
        if isinstance(message, MetricMessage):
            # logger.debug(f"ON_EXCHANGE:{message}")
            self.update(message)
        elif isinstance(message, OutputRequestMessage):
            logger.debug(f"ON_EXCHANGE:{message}")
            self.update(message)


    def has_divergence(self, message: GuidanceMessage) -> Tuple[bool, int]:
        """Checks if message has divergence with current path.

        Args:
            message: Incoming message.

        Returns:
            Tuple of (has diverged, shared ancestor index). Index will be -1 if no divergence.

        Raises:
            Exception if there is no shared ancestor (including root). This should not happen.
        """
        if not isinstance(message, TraceMessage):
            return False, -1

        # If we diverge from the model path, truncate and reset
        message_trace_node = self._trace_handler[message.trace_id]

        prev_trace_messages = [x for x in self._messages if isinstance(x, TraceMessage)]
        trace_messages_len = len(prev_trace_messages)
        if trace_messages_len == 0:
            return False, -1
        elif trace_messages_len == 1:
            if isinstance(self._messages[0], TraceMessage):
                last_trace_node = self._trace_handler[self._messages[0].trace_id]
                if message_trace_node.parent == last_trace_node:
                    return False, -1
                else:
                    return True, 0
            else:
                return False, -1
        else:
            last_trace_message = prev_trace_messages[-1]
            last_trace_node = self._trace_handler[last_trace_message.trace_id]

            if last_trace_node not in message_trace_node.path():
                logger.debug(f"DIVERGENCE:curr:{message_trace_node}")
                logger.debug(f"DIVERGENCE:prev:{last_trace_node}")

                # Truncate path that is no longer used by current trace node
                ancestor_idx = -1
                ancestors = set(message_trace_node.ancestors())
                for idx, prev_message in enumerate(self._messages):
                    if isinstance(prev_message, TraceMessage):
                        prev_trace_node = self._trace_handler[prev_message.trace_id]
                        if prev_trace_node in ancestors:
                            ancestor_idx = idx
                if ancestor_idx == -1:
                    if message_trace_node.parent == last_trace_node.root():  # pragma: no cover
                        ancestor_idx = 0
                    else:
                        logger.debug(f"DIVERGENCE:full_reset")
                        ancestor_idx = 0

                return True, ancestor_idx
            else:
                return False, -1


    def update(self, message: GuidanceMessage, topic=DEFAULT_TOPIC) -> None:
        from ..registry import get_exchange
        from ..registry import get_bg_async

        out_messages = []

        # Metrics
        if isinstance(message, MetricMessage):
            if self._running:
                # logger.debug(f"RENDERER:metric:{message}")
                pass
            else:
                return

        if isinstance(message, ExecutionCompletedMessage):
            # Execution completed
            logger.debug("RENDERER:execution end")
            self._completed = True
            self._running = False
            get_exchange().publish(message)

            if message.is_err:
                out_messages.append(MetricMessage(name="status", value="Error"))
            else:
                out_messages.append(MetricMessage(name="status", value="Done"))
        elif not self._running and isinstance(message, TraceMessage):
            # Execution started
            logger.debug("RENDERER:execution start")
            started_msg = ExecutionStartedMessage()
            out_messages.append(started_msg)
            out_messages.append(MetricMessage(name="status", value='Running'))

            _, self.last_cell_session_id = ipy_handle_event_once(
                partial(_on_cell_completion, weakref.ref(self)),
                'post_run_cell'
            )
            self._new_widget_needed = True
            self._running = True
            self._completed = False

            # TODO(nopdive): Fire off execution immediately to renderer subscribers. Review later.
            get_bg_async().call_soon_threadsafe(self.recv_queue.put_nowait, serialize_message(started_msg))

        # Check if message has diverged from prev messages
        diverged, shared_ancestor_idx = self.has_divergence(message)
        if diverged:
            logger.debug("RENDERER:diverged")
            out_messages.append(ResetDisplayMessage())
            out_messages[len(out_messages):] = self._messages[:shared_ancestor_idx]
            self._messages.clear()
        
        # Check if requested to reset and replay
        if isinstance(message, OutputRequestMessage):
            logger.debug(f"RENDERER:replay:{message}")
            out_messages.append(ResetDisplayMessage())
            out_messages[len(out_messages):] = self._messages[:]

        # Reset if needed
        if self._new_widget_needed:
            logger.debug("RENDERER:new widget needed")

            # Clear messages
            self._messages = []

            if self.stitch_widget is not None and self.stitch_widget_observed:
                self.stitch_widget.unobserve(self._stitch_on_clientmsg, names='clientmsg')
                self.stitch_widget_observed = False
                logger.debug("RENDERER:widget unobserved (new)")

            self.stitch_widget = _create_stitch_widget()
            self._stitch_on_clientmsg = partial(_on_stitch_clientmsg, weakref.ref(self.recv_queue))
            self.stitch_widget.observe(self._stitch_on_clientmsg, names='clientmsg')
            self.stitch_widget_observed = True
            logger.debug("RENDERER:widget observed (new)")

            # Redraw
            display(self.stitch_widget)

            self._new_widget_needed = False

        # Append current message to outgoing
        out_messages.append(message)

        # Send outgoing messages to client
        for out_message in out_messages:
            if isinstance(out_message, TraceMessage):
                # logger.debug(f"RENDERER:out:{out_message}")
                self.last_trace_id = out_message.trace_id

            self._messages.append(out_message)
            get_bg_async().call_soon_threadsafe(self.send_queue.put_nowait, out_message)


class DoNothingRenderer(Renderer):
    """ It does nothing."""

    def __init__(self, trace_handler: TraceHandler) -> None:
        """ Initializes.

        Args:
            trace_handler: Trace handler of an engine.
        """
        self._trace_handler = trace_handler
        super().__init__()

    def update(self, message: GuidanceMessage, topic: str = DEFAULT_TOPIC) -> None:
        pass


class AutoRenderer(Renderer):
    """ Automatically detects which renderer to use based on environment."""

    def __init__(self, trace_handler: TraceHandler):
        """ Initializes.

        Args:
            trace_handler: Trace handler of an engine.
        """
        self._env = Environment()

        if self._env.is_notebook():
            if stitch_installed:
                self._renderer = JupyterWidgetRenderer(trace_handler=trace_handler)
            else:
                self._renderer = DoNothingRenderer(trace_handler=trace_handler)
        elif self._env.is_terminal():
            # TODO(nopdive): When IPython events are figured out (cell completion)
            #                hook up terminal interface separate to non-interactive
            #                shell.
            self._renderer = DoNothingRenderer(trace_handler=trace_handler)
        else:  # pragma: no cover
            logger.error("Env detection has failed. This is a bug.")
            warn("Env detection has failed. No renderer will be provided.")
            self._renderer = DoNothingRenderer(trace_handler=trace_handler)

        super().__init__()

    def update(self, message: GuidanceMessage, topic: str = DEFAULT_TOPIC) -> None:
        self._renderer.update(message, topic=topic)

    def renderer_type(self) -> type:
        """Type of renderer that has been selected based on environment.

        Returns:
            Type of selected renderer.
        """
        return type(self._renderer)


""" Jupyter specific utilities."""


from typing import Callable, Any, Tuple, Optional
import logging
from uuid import uuid4

try:
    from IPython import get_ipython
except ImportError:
    pass

logger = logging.getLogger(__name__)

IPythonCallback = Callable[[Any], None]


def ipy_handle_event_once(cb: IPythonCallback, event_name: str) -> Tuple[Optional[IPythonCallback], str]:
    ipy = get_ipython()
    cell_session_id = str(uuid4())

    if ipy is None:
        return None, ""
    
    def cb_closure(msg):
        cb(info=msg)
        ipy.events.unregister(event_name, cb_closure)
    ipy.events.register(event_name, cb_closure)
    
    return cb_closure, cell_session_id


import pathlib
import typing
import pkg_resources
import base64

from .._guidance import guidance
from .._utils import bytes_from
# from ..trace._trace import VideoInput
from ..trace._trace import VideoOutput


@guidance
def video(lm, src: typing.Union[str, pathlib.Path, bytes], allow_local: bool = True):
    # TODO(nopdive): Mock for testing. Remove all of this code later.
    bytes_data = bytes_from(src, allow_local=allow_local)
    base64_string = base64.b64encode(bytes_data).decode('utf-8')
    lm += VideoOutput(value=base64_string, is_input=True)
    # lm += VideoInput(value=base64_string)
    return lm


@guidance
def gen_video(lm):
    # TODO(nopdive): Mock for testing. Remove all of this code later.
    with open(
        pkg_resources.resource_filename("guidance", "resources/sample_video.mp4"), "rb"
    ) as f:
        bytes_data = f.read()
        base64_string = base64.b64encode(bytes_data).decode('utf-8')
        lm += VideoOutput(value=base64_string, is_input=False)
    return lm



"""Trace tree of inputs & outputs generated from a guidance program.

The first implementation aims for simplicity.
Once benchmark figures are out, we'll figure out what to optimize.

The most critical class is the trace handler. See its documentation for trace design & motivations.
"""

from ._trace import NodeAttr, InputAttr, OutputAttr, StatefulGuidanceInput, StatelessGuidanceInput
from ._trace import LiteralInput, EmbeddedInput, ImageInput, RoleCloserInput, RoleOpenerInput
from ._trace import TextOutput, ImageOutput, CaptureOutput, TraceNode, TraceHandler



from ._client import Client
from ._model import Model
from ._state import State

__all__ = [
    "Model",
    "role",
    "State",
    "Message",
    "Client",
    "ASTNode",
    "ContentChunk",
    "MessageChunk",
]



__version__ = "0.2.1"

import sys
import types

from . import models
from ._guidance import guidance

from ._ast import GrammarNode, Function
from ._utils import strip_multiline_string_indents


# This makes the guidance module callable
class _Guidance(types.ModuleType):
    def __call__(
        self, f=None, *, stateless=False, cache=None, dedent=True, model=models.Model
    ):
        return guidance(
            f, stateless=stateless, cache=cache, dedent=dedent, model=model
        )


sys.modules[__name__].__class__ = _Guidance

# we expose all the library functions at the top level of the module
from .library import *


import atexit
import logging
import operator
import os
import sys
from itertools import takewhile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Sequence
import ctypes

import numpy as np

from .._schema import GenToken, GenTokenExtra
from .._utils import normalize_notebook_stdout_stderr, softmax
from ._base import Model
from ._engine import Engine, EngineClient, EngineState, Tokenizer

try:
    import llama_cpp

    is_llama_cpp = True
except ModuleNotFoundError:
    is_llama_cpp = False

if TYPE_CHECKING:
    from llama_cpp.llama_tokenizer import LlamaTokenizer
    from llama_cpp.llama import Llama
    from llama_cpp.llama_chat_format import Jinja2ChatFormatter

logger = logging.getLogger(__name__)

shutdown_none = True


def set_shutdown_flag():
    global shutdown_none
    # python can set anything to None at shutdown, so use None
    shutdown_none = None


atexit.register(set_shutdown_flag)


class _LlamaBatchContext:
    def __init__(self, n_batch, n_ctx):
        self._llama_batch_free = llama_cpp.llama_batch_free
        self.batch = llama_cpp.llama_batch_init(n_batch, 0, n_ctx)
        if self.batch is None:
            raise Exception("call to llama_cpp.llama_batch_init returned NULL.")

    def __del__(self):
        if shutdown_none is not None:
            llama_batch_free = getattr(self, "_llama_batch_free", None)
            batch = getattr(self, "batch", None)
            if batch is not None and llama_batch_free is not None:
                self._llama_batch_free = None
                self.batch = None
                llama_batch_free(batch)

def detokenize(tokenizer: "LlamaTokenizer", tokens: list[int], special: bool, size: int) -> bytes:
    """Re-implementation of llama_cpp.LLamaTokenizer.detokenize that ditches the hard-coded size=32"""
    output = b""
    buffer = ctypes.create_string_buffer(size)
    for token in tokens:
        n = llama_cpp.llama_token_to_piece(
            tokenizer._model.vocab,
            llama_cpp.llama_token(token),
            buffer,
            size,
            0,
            special
        )
        if n < 0:
            raise ValueError(f"Error writing token {token} to buffer of size {size}. Error: {n}")
        assert n <= size
        output += bytes(buffer[:n])
    # NOTE: Llama1 models automatically added a space at the start of the prompt
    # this line removes a leading space if the first token is a beginning of sentence token
    return (
        output[1:]
        if len(tokens) > 0 and tokens[0] == tokenizer._model.token_bos() and output[0:1] == b" "
        else output
        )

class LlamaCppTokenizer(Tokenizer):
    def __init__(self, model_obj, chat_template=None):
        self._model_obj = model_obj

        self._sentinel_bytes = "\x02".encode()
        self._sentinel_tokens = self._model_obj.tokenize(
            self._sentinel_bytes, add_bos=False, special=True
        )

        tokenizer = llama_cpp.LlamaTokenizer(model_obj)
        vocab = llama_cpp.llama_model_get_vocab(model_obj.model)
        if vocab is None:
            raise Exception("call to llama_cpp.llama_model_get_vocab returned NULL.")

        if not hasattr(tokenizer, "llama"):
            tokenizer.llama = tokenizer._model

        # get the bytes strings for all the tokens
        tokens = []
        special_token_ids = []
        for i in range(tokenizer.llama.n_vocab()):
            tok_attrs = tokenizer.llama.token_get_attr(i)
            if tok_attrs & llama_cpp.LLAMA_TOKEN_ATTR_CONTROL:
                special_token_ids.append(i)
            tok = detokenize(tokenizer, [i], special=True, size=256)
            tokens.append(tok)

        # Chat Template logic
        if chat_template is None:
            if (
                hasattr(self._model_obj, "metadata")
                and "tokenizer.chat_template" in self._model_obj.metadata
            ):
                chat_template = self._model_obj.metadata["tokenizer.chat_template"]

        super().__init__(
            tokens, chat_template, tokenizer.llama.token_bos(), tokenizer.llama.token_eos(), special_token_ids
        )

    def encode(self, byte_string: bytes) -> list[int]:
        # Workaround for the LlamaCpp prepending spaces on encoding
        raw_tokens = self._model_obj.tokenize(
            self._sentinel_bytes + byte_string, add_bos=False, special=True
        )
        assert raw_tokens[: len(self._sentinel_tokens)] == self._sentinel_tokens
        return raw_tokens[len(self._sentinel_tokens) :]


class LlamaCppEngine(Engine):
    """The core class that runs inference using llama.cpp."""

    def __init__(
        self,
        model,
        compute_log_probs,
        chat_template=None,
        enable_backtrack=True,
        enable_ff_tokens=True,
        enable_monitoring=True,
        **kwargs,
    ):
        if not is_llama_cpp:
            raise Exception(
                "Please install llama-cpp-python with `pip install llama-cpp-python` in order to use guidance.models.LlamaCpp!"
            )

        if isinstance(model, Path):
            model = str(model)
        if model is None or isinstance(model, str) and len(model.strip()) == 0:
            model = os.environ.get("LLAMA_CPP_MODEL", "")
            if len(model.strip()) == 0:
                try:
                    with open(os.path.expanduser("~/.llama_cpp_model"), "r") as file:
                        model = file.read().replace("\n", "")
                except:
                    pass
                if len(model.strip()) == 0:
                    raise ValueError(
                        "If model is None then a model file must be specified in either the LLAMA_CPP_MODEL environment variable or in the ~/.llama_cpp_model file."
                    )

        if isinstance(model, str):
            self.model = model
            if "verbose" not in kwargs:
                kwargs["verbose"] = False

            # patch over https://github.com/abetlen/llama-cpp-python/issues/729
            try:
                sys.stdout.fileno()
            except:
                logger.warn(
                    "Cannot use verbose=True in this context (probably CoLab). See https://github.com/abetlen/llama-cpp-python/issues/729"
                )
                kwargs["verbose"] = True  # llama-cpp-python can't hide output in this case

            with normalize_notebook_stdout_stderr():
                self.model_obj = llama_cpp.Llama(model_path=model, logits_all=True, **kwargs)
        elif isinstance(model, llama_cpp.Llama):
            self.model = model.__class__.__name__
            self.model_obj = model
        else:
            raise TypeError("model must be None, a file path string, or a llama_cpp.Llama object.")

        self._context = _LlamaBatchContext(self.model_obj.n_batch, self.model_obj.n_ctx())
        self._cache_token_ids = []

        super().__init__(LlamaCppTokenizer(self.model_obj, chat_template=chat_template),
                         compute_log_probs=compute_log_probs, enable_backtrack=enable_backtrack,
                         enable_ff_tokens=enable_ff_tokens, enable_monitoring=enable_monitoring, **kwargs)

        self._n_vocab = len(self.tokenizer.tokens)

    def get_logits(self, token_ids):
        """Computes the logits for the given token state.

        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        """

        if len(token_ids) == 0:
            raise ValueError("token_ids must contain some tokens.")

        # check what we have already cached
        cache_token_ids = self._cache_token_ids
        num_cached = sum(takewhile(operator.truth, map(operator.eq, token_ids, cache_token_ids)))
        if num_cached == len(token_ids):
            if num_cached == len(cache_token_ids):
                return self._cached_logits
            num_cached = (
                num_cached - 1
            )  # llama_cpp doesn't like it when we pass in 0 new tokens, so re-input one

        # make sure we don't run off the end of the model's context
        if self.model_obj.n_ctx() <= len(token_ids):
            raise Exception(
                f"Attempted to use a context length of {len(token_ids)} tokens, but this LlamaCpp model is only configured to support up to {self.model_obj.n_ctx()}!"
            )

        self._cache_token_ids = token_ids.copy()

        # clear obsolete parts of kv cache
        llama_cpp.llama_kv_cache_seq_rm(self.model_obj.ctx, -1, num_cached, -1)

        # eval the model
        n_batch = self.model_obj.n_batch
        batch = self._context.batch
        for i in range(num_cached, len(token_ids), n_batch):
            n_tokens = min(i + n_batch, len(token_ids)) - i
            batch.n_tokens = n_tokens
            for j in range(n_tokens):
                batch.token[j] = token_ids[i + j]
                batch.pos[j] = i + j
                batch.seq_id[j][0] = 0
                batch.n_seq_id[j] = 1
                batch.logits[j] = False

            if i + n_tokens == len(token_ids):
                batch.logits[n_tokens - 1] = True

            ret = llama_cpp.llama_decode(self.model_obj.ctx, batch)
            self.metrics.engine_input_tokens += n_tokens
            if ret != 0:
                raise Exception(f"Call to llama_cpp.llama_decode returned {ret}.")

        self.metrics.engine_output_tokens += 1

        # get the logits
        logits = llama_cpp.llama_get_logits(self.model_obj.ctx)
        if llama_cpp.__version__ < "0.2.58":
            logits = logits[(n_tokens - 1) * self._n_vocab : n_tokens * self._n_vocab]
        logits = np.ctypeslib.as_array(logits, shape=(self._n_vocab,)).copy()

        self._cached_logits = logits

        return logits

    def get_per_token_topk_probs(
        self, token_ids: list[int], top_k: int = 5
    ) -> list[GenTokenExtra]:
        if len(token_ids) == 0:
            return []

        # clear kv-cache
        llama_cpp.llama_kv_cache_seq_rm(self.model_obj.ctx, -1, 0, -1)

        # create new context
        context = self._context
        n_batch = self.model_obj.n_batch
        batch = context.batch

        logits_batch = []
        for i in range(0, len(token_ids), n_batch):
            n_tokens = min(i + n_batch, len(token_ids)) - i
            batch.n_tokens = n_tokens

            for j in range(n_tokens):
                batch.token[j] = token_ids[i + j]
                batch.pos[j] = i + j
                batch.seq_id[j][0] = 0
                batch.n_seq_id[j] = 1
                batch.logits[j] = True

            ret = llama_cpp.llama_decode(self.model_obj.ctx, batch)
            if ret != 0:
                raise Exception(f"Call to llama_cpp.llama_decode returned {ret}.")

            # get all the logits
            if llama_cpp.__version__ < "0.2.58":
                logits = llama_cpp.llama_get_logits(self.model_obj.ctx)
                logits = logits.reshape((-1, self._n_vocab))
                logits_batch.extend(logits)
            else:
                for j in range(n_tokens):
                    logits = llama_cpp.llama_get_logits_ith(self.model_obj.ctx, j)
                    logits = np.ctypeslib.as_array(logits, shape=(self._n_vocab,)).copy()
                    logits_batch.append(logits)

        # process each token's logits
        text_sequence = []

        # add 1st token
        _bytes = self.tokenizer.decode([token_ids[0]])
        try:
            _text = _bytes.decode("utf-8")
        except Exception as e:
            _text = str(_bytes)
            print(f"Failed to decode token: {token_ids[0]}, error: {e}, _bytes: {str(_bytes)}")
        text_sequence.append(
            GenTokenExtra(
                token_id=token_ids[0],
                prob=1.0,
                text=_text,
                top_k=[GenToken(token_id=token_ids[0], prob=1.0, text=_text)],
            )
        )

        for token_id, logits in zip(token_ids[1:], logits_batch[:-1]):
            _probs = softmax(logits)

            # get the top k indices
            top_k_ids, top_k_probs = self._get_top_k(_probs.copy(), top_k, ascending=False)
            if token_id not in top_k_ids:
                top_k_ids = np.append(top_k_ids, token_id)
                top_k_probs = np.append(top_k_probs, _probs[token_id])

            top_k_list = []
            for _token_id, _prob in zip(top_k_ids, top_k_probs):
                _text = ""
                try:
                    _text = self.tokenizer.decode([_token_id]).decode("utf-8")
                except Exception as e:
                    _bytes = self.tokenizer.decode([_token_id])
                    _text = str(_bytes)
                    print(
                        f"Failed to decode token: {_token_id}, error: {e}, _bytes: {str(_bytes)}"
                    )
                top_k_list.append(GenToken(token_id=_token_id, prob=_prob, text=_text))

            text_sequence.append(
                GenTokenExtra(
                    token_id=token_id,
                    prob=_probs[token_id],
                    text=self.tokenizer.decode([token_id]).decode("utf-8"),
                    top_k=top_k_list,
                )
            )

        return text_sequence

    def _get_top_k(self, probs: np.ndarray, k: int, axis: int = None, ascending: bool = True):
        if not ascending:
            probs *= -1
        ind = np.argpartition(probs, k, axis=axis)
        ind = np.take(ind, np.arange(k), axis=axis)  # k non-sorted indices
        probs = np.take_along_axis(probs, ind, axis=axis)  # k non-sorted values

        # sort within k elements
        ind_part = np.argsort(probs, axis=axis)
        ind = np.take_along_axis(ind, ind_part, axis=axis)
        if not ascending:
            probs *= -1
        val = np.take_along_axis(probs, ind_part, axis=axis)
        return ind, val


class LlamaCpp(Model):
    def __init__(
        self,
        model=None,
        echo=True,
        compute_log_probs=False,
        api_key=None,
        chat_template=None,
        enable_backtrack=True,
        enable_ff_tokens=True,
        enable_monitoring=True,
        **llama_cpp_kwargs,
    ):
        """Build a new LlamaCpp model object that represents a model in a given state."""

        engine = LlamaCppEngine(
            model,
            compute_log_probs=compute_log_probs,
            chat_template=chat_template,
            enable_backtrack=enable_backtrack,
            enable_ff_tokens=enable_ff_tokens,
            enable_monitoring=enable_monitoring,
            **llama_cpp_kwargs,
        )
        state = EngineState()
        client = EngineClient(engine)
        super().__init__(client=client, state=state, echo=echo)



# TODO(nopdive): This module requires a memory review.

import queue
import threading
from contextvars import ContextVar, copy_context
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Generic, Iterator, Optional, TypeVar, Union

from typing_extensions import Self

from ..._ast import ASTNode, Function, _parse_tags
from ..._ast import (
    ASTNode,
    Function,
    GenAudio,
    ImageBlob,
    ImageUrl,
    LiteralNode,
    RoleEnd,
    RoleStart,
    _parse_tags,
)
from ...trace import (
    ImageInput,
    LiteralInput,
    NodeAttr,
    RoleCloserInput,
    RoleOpenerInput,
    StatelessGuidanceInput,
    TextOutput,
    TraceNode,
)
from ...trace._trace import AudioInput
from ...visual import TraceMessage
from ._client import Client
from ._state import State

if TYPE_CHECKING:
    from ...library._block import Block

_active_blocks: ContextVar[tuple["Block", ...]] = ContextVar("active_blocks", default=())
_event_queues: ContextVar[tuple[queue.Queue["Model"], ...]] = ContextVar(
    "event_queues", default=()
)
_id_counter: int = 0


def _gen_id():
    global _id_counter

    _id = _id_counter
    _id_counter += 1

    return _id


S = TypeVar("S", bound=State)
D = TypeVar("D", bound=Any)


class Model(Generic[S]):
    def __init__(
        self,
        client: Client[S],
        state: S,
        echo: bool = True,
    ) -> None:
        self.echo = echo
        self._client = client
        self._state = state
        self._active_blocks: dict[Block, int] = {}
        self.token_count: int = 0

        self._parent: Optional["Model"] = None
        self._parent_id: Optional[int] = None
        self._id: int = _gen_id()
        self._trace_nodes: set[TraceNode] = set()
        self._update_trace_node(self._id, self._parent_id, None)

    def _update_trace_node(
        self, identifier: int, parent_id: Optional[int], node_attr: Optional[NodeAttr] = None
    ) -> None:
        from ...registry import get_trace_handler, get_renderer

        trace_handler = get_trace_handler()
        trace_node = trace_handler.update_node(identifier, parent_id, node_attr)
        self._trace_nodes.add(trace_node)
        if self.echo:
            get_renderer().update(
                TraceMessage(
                    trace_id=identifier,
                    parent_trace_id=parent_id,
                    node_attr=node_attr,
                ),
            )

    def __add__(self, other: Union[str, Function, ASTNode]) -> Self:
        self = self._apply_blocks()
        if isinstance(other, str):
            if other == "":
                return self
            other = _parse_tags(other)
        if isinstance(other, Function):
            return other(self)
        if isinstance(other, ASTNode):
            self = self._apply_node(other)
            self = self._update_open_block_captures()
            return self
        return NotImplemented

    def _apply_node(self, node: ASTNode) -> Self:
        self = self.copy()

        # Input side of trace handler.
        # TODO: StatefulGuidanceInput up in __add__?
        if isinstance(node, RoleStart):
            self._update_trace_node(self._id, self._parent_id, RoleOpenerInput(name=node.role))
        elif isinstance(node, RoleEnd):
            self._update_trace_node(self._id, self._parent_id, RoleCloserInput(name=node.role))
        elif isinstance(node, LiteralNode):
            self._update_trace_node(self._id, self._parent_id, LiteralInput(value=node.value))
        elif isinstance(node, ImageBlob):
            self._update_trace_node(self._id, self._parent_id, ImageInput(value=node.data))
        elif isinstance(node, ImageUrl):
            # TODO -- let's avoid downloading it here
            pass
        elif isinstance(node, GenAudio):
            self._update_trace_node(
                self._id, self._parent_id, AudioInput(value="")
            )  # TODO -- what goes here?
        else:
            self._update_trace_node(self._id, self._parent_id, StatelessGuidanceInput(value=node))

        for i, output_attr in enumerate(self._client.run(self._state, node)):
            if isinstance(output_attr, TextOutput):
                # TODO: put this elsewhere (inside state?)
                self.token_count += output_attr.token_count
            if i != 0:
                # On the first iteration, we already have a fresh trace node
                # TODO: should be allowed to associate multiple output_attrs with a single input node?
                # TODO: put this responsibility on the client in the case that it breaks a single input
                # node into multiple input nodes to be handled sequentially?
                self._parent_id = self._id
                self._id = _gen_id()
            self._update_trace_node(self._id, self._parent_id, output_attr)
            # Stream current model state
            self._send_to_event_queue()
        return self

    def _send_to_event_queue(self) -> None:
        """For streaming"""
        for event_queue in _event_queues.get():
            event_queue.put(self.copy())

    def stream(self) -> "ModelStream":
        """Return a new model stream object that delays execution until it is iterated over."""
        return ModelStream(self)

    def _apply_blocks(self) -> Self:
        self = self.copy()
        global_active_blocks = _active_blocks.get()
        for block, start_index in list(reversed(self._active_blocks.items())):
            # Close blocks that are not globally active anymore
            if block not in global_active_blocks:
                self._active_blocks.pop(block)
                if block.closer is not None:
                    closer = block.closer
                    if isinstance(closer, str):
                        closer = _parse_tags(closer)
                    if isinstance(closer, Function):
                        raise NotImplementedError(
                            "Stateful block opener/closer functions are not yet supported"
                        )
                    self = self._apply_node(closer)
            # Update capture regardless of whether or not it's been closed
            if block.name is not None:
                self = self.set(block.name, str(self)[start_index:])
        for block in global_active_blocks:
            # Open blocks that are not yet locally active
            if block not in self._active_blocks:
                # Set start_index to the current length
                self._active_blocks[block] = len(self)
                if block.opener is not None:
                    opener = block.opener
                    if isinstance(opener, str):
                        opener = _parse_tags(opener)
                    if isinstance(opener, Function):
                        raise NotImplementedError(
                            "Stateful block opener/closer functions are not yet supported"
                        )
                    self = self._apply_node(opener)
        return self

    def _update_open_block_captures(self) -> Self:
        self = self.copy()
        for block, start_index in self._active_blocks.items():
            if block.name is not None:
                self = self.set(block.name, str(self)[start_index:])
        return self

    def copy(self) -> Self:
        obj = object.__new__(self.__class__)
        obj.__dict__.update(self.__dict__)

        obj._state = deepcopy(self._state)
        obj._active_blocks = {**self._active_blocks}
        obj._id = _gen_id()
        obj._parent_id = self._id
        obj._trace_nodes = set()
        obj._parent = self
        obj._update_trace_node(obj._id, obj._parent_id, None)
        return obj

    def __str__(self) -> str:
        return str(self._state)

    def __len__(self):
        return len(str(self))

    def __setitem__(self, key, value):
        raise Exception(
            "Model objects are immutable so you can't use __setitem__! Consider using the .set(key, value) method instead to create a new updated model object."
        )

    def __getitem__(self, key: str) -> Any:
        try:
            captures = self._state.captures[key]
        except KeyError:
            raise KeyError(f"Model does not contain the variable '{key}'")
        if isinstance(captures, list):
            return [c["value"] for c in captures]
        else:
            return captures["value"]

    def __contains__(self, key: str) -> bool:
        return key in self._state.captures

    def get(self, key: str, default: Optional[D] = None) -> Union[str, list[str], None, D]:
        """Return the value of a variable, or a default value if the variable is not present.

        Parameters
        ----------
        key : str
            The name of the variable.
        default : Any
            The value to return if the variable is not current set.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def set(self, key: str, value: Union[str, list[str]]) -> Self:
        """Return a new model with the given variable value set.

        Parameters
        ----------
        key : str
            The name of the variable to be set.
        value : str
            The value to set the variable to.
        """
        self = self.copy()
        if isinstance(value, list):
            self._state.captures[key] = [{"value": v, "log_prob": None} for v in value]
        else:
            self._state.captures[key] = {"value": value, "log_prob": None}
        return self

    def remove(self, key: str) -> Self:
        """Return a new model with the given variable deleted.

        Parameters
        ----------
        key : str
            The variable name to remove.
        """
        self = self.copy()
        self._state.captures.pop(key)
        return self

    def log_prob(
        self, key: str, default: Optional[D] = None
    ) -> Union[float, list[Union[float, None]], None, D]:
        """Return the log probability of a variable, or a default value if the variable is not present.

        Parameters
        ----------
        key : str
            The name of the variable.
        default : Any
            The value to return if the variable is not current set.
        """
        try:
            captures = self._state.captures[key]
        except KeyError:
            return default
        if isinstance(captures, list):
            return [c["log_prob"] for c in captures]
        else:
            return captures["log_prob"]

    def __getattribute__(self, name):
        if name == "engine":
            # For legacy model.engine access (mostly for tests...)
            return getattr(self._client, "engine")
        return super().__getattribute__(name)


class ModelStream:
    def __init__(
        self,
        model: Model,
        grammar: Union["ModelStream", str, ASTNode, Function, None] = None,
        timeout=5,
    ) -> None:
        """Create a model stream object that delays execution until it is iterated over."""
        if model.echo:
            model = model.copy()
            model.echo = False  # turn off display echoing
        self.model = model
        self.grammar = grammar
        self.timeout = timeout

    def __add__(self, grammar: Union[str, ASTNode]) -> Self:
        """Extend this delayed chain of execution with another grammar append."""
        if self.grammar is None:
            return ModelStream(self.model, grammar)
        else:
            return ModelStream(self.model, self.grammar + grammar)

    def _inner_run(self, model):
        """This runs the model stream without iterating, and is only using internally by __iter__."""
        if isinstance(self.grammar, ModelStream):
            model = self.grammar._inner_run(model)
        elif self.grammar is None:
            model = self.model + ""
        else:
            model = self.model + self.grammar

    def __iter__(self) -> Iterator[Model]:
        """Starts a thread to execute the model and grammar, yielding events as they occur."""

        events = queue.Queue()
        event_queues = _event_queues.get() + (events,)
        token = _event_queues.set(event_queues)

        # Define the target function for the thread
        def target(ctx):
            _event_queues.set(ctx[_event_queues])
            try:
                self._inner_run(self.model)
                events.put(None)  # mark that we are done
            except BaseException as ex:
                events.put(ex)

        # Start the thread
        thread = threading.Thread(target=target, args=(copy_context(),))
        thread.start()

        # Yield events from the queue as they become available
        while True:
            try:
                # Wait for an event with a timeout to allow for thread termination
                event = events.get(timeout=self.timeout)
                if event is None:
                    break
                elif isinstance(event, BaseException):
                    raise event
                yield event
            except queue.Empty:
                # Check if the thread is still alive
                if not thread.is_alive():
                    break

        # Ensure the thread has completed
        thread.join()

        # Reset the event queues context variable
        _event_queues.reset(token)



from ._tokenizer import Tokenizer  # isort:skip
from ._client import EngineClient, Llama3VisionClient, Phi3VisionClient
from ._engine import Engine
from ._state import EngineState

__all__ = [
    "Tokenizer",
    "Engine",
    "EngineClient",
    "EngineState",
    "Llama3VisionClient",
    "Phi3VisionClient",
]



import numpy as np
from ._engine import Tokenizer
from ..chat import load_template_class
from typing import List

class ByteTokenizer(Tokenizer):
    def __init__(self, chat_template=None):
        # directly map integer values to byte strings
        all_bytes = [bytes([i]) for i in range(256)]
        bos = b"<s>"
        tokens = np.array(all_bytes + [bos], dtype="object")
        chat_template = load_template_class(chat_template)
        super().__init__(tokens, chat_template, bos_token_id=256)

    def encode(self, byte_string: bytes) -> List[int]:
        """Returns a list of tokens that represent the given byte string."""
        if isinstance(byte_string, str):
            byte_string = byte_string.encode("utf8")
        i = 0
        result = []
        while i < len(byte_string):
            if byte_string[i:i+3] == b'<s>':
                result.append(256)
                i += 3  # Skip the next two characters as part of '<s>'
            else:
                result.append(byte_string[i])
                i += 1
        return result



import base64
from typing import Generic, Iterator, TypeVar

from ..._ast import (
    ASTNode,
    AudioBlob,
    GenAudio,
    GrammarNode,
    ImageBlob,
    ImageUrl,
    JoinNode,
    JsonNode,
    LarkNode,
    LiteralNode,
    RegexNode,
    RepeatNode,
    RoleEnd,
    RoleStart,
    RuleNode,
    SelectNode,
    SubgrammarNode,
    SubstringNode,
)
from ..._utils import bytes_from
from ...trace import OutputAttr
from ._state import State

S = TypeVar("S", bound=State)


class Client(Generic[S]):
    def run(self, state: S, node: ASTNode, **kwargs) -> Iterator[OutputAttr]:
        yield from node.simplify()._run(self, state, **kwargs)

    def _role_start(self, state: S, node: RoleStart, **kwargs) -> Iterator[OutputAttr]:
        if state.active_role is not None:
            raise ValueError(
                f"Cannot open role {node.role!r}: {state.active_role!r} is already open."
            )
        return self.role_start(state, node, **kwargs)

    def role_start(self, state: S, node: RoleStart, **kwargs) -> Iterator[OutputAttr]:
        raise UnsupportedNodeError(client=self, node=node)

    def _role_end(self, state: S, node: RoleEnd, **kwargs) -> Iterator[OutputAttr]:
        if state.active_role is None:
            raise ValueError("Cannot close role without active role")
        if state.active_role != node.role:
            raise ValueError(f"Cannot close role {node.role!r}: {state.active_role!r} is open.")
        return self.role_end(state, node, **kwargs)

    def role_end(self, state: S, node: RoleEnd, **kwargs) -> Iterator[OutputAttr]:
        raise UnsupportedNodeError(client=self, node=node)

    def text(self, state: S, node: LiteralNode, **kwargs) -> Iterator[OutputAttr]:
        raise UnsupportedNodeError(client=self, node=node)

    def image_blob(self, state: S, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        raise UnsupportedNodeError(client=self, node=node)

    def image_url(self, state: S, node: ImageUrl, **kwargs) -> Iterator[OutputAttr]:
        image_bytes = bytes_from(node.url, allow_local=False)
        base64_string = base64.b64encode(image_bytes).decode("utf-8")
        return self.image_blob(state, ImageBlob(data=base64_string), **kwargs)

    def grammar(self, state: S, node: GrammarNode, **kwargs) -> Iterator[OutputAttr]:
        raise UnsupportedNodeError(client=self, node=node)

    def regex(self, state: S, node: RegexNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def select(self, state: S, node: SelectNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def join(self, state: S, node: JoinNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def repeat(self, state: S, node: RepeatNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def substring(self, state: S, node: SubstringNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def rule(self, state: S, node: RuleNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def subgrammar(self, state: S, node: SubgrammarNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def json(self, state: S, node: JsonNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def lark(self, state: S, node: LarkNode, **kwargs) -> Iterator[OutputAttr]:
        return self.grammar(state, node, **kwargs)

    def audio_blob(self, state: S, node: AudioBlob, **kwargs) -> Iterator[OutputAttr]:
        raise UnsupportedNodeError(client=self, node=node)

    def gen_audio(self, state: S, node: GenAudio, **kwargs) -> Iterator[OutputAttr]:
        raise UnsupportedNodeError(client=self, node=node)


class UnsupportedNodeError(ValueError):
    def __init__(self, client: Client, node: ASTNode):
        super().__init__(f"{client} does not support {node!r} of type {type(node)}")
        self.client = client
        self.node = node



# NOTE(nopdive): Consider moving singleton factories to registry static class.

import threading

from .._schema import GuidanceEngineMetrics
from ..metrics import Monitor, PeriodicMetricsGenerator
from ..trace import TraceHandler
from ..visual import AutoRenderer, Renderer, TopicExchange
from .._bg import BackgroundAsync

_monitor_lock = threading.Lock()
_monitor = None
_periodic_metrics_gen = None

_bg_async_lock = threading.Lock()
_bg_async = None

_exchange_lock = threading.Lock()
_exchange = None

_trace_handler_lock = threading.Lock()
_trace_handler = None

_renderer_lock = threading.Lock()
_renderer = None


def get_monitor() -> Monitor:
    global _monitor
    global _monitor_lock
    global _periodic_metrics_gen

    with _monitor_lock:
        if _monitor is None:
            _monitor = Monitor(GuidanceEngineMetrics())
            _monitor.start()
            _periodic_metrics_gen = PeriodicMetricsGenerator(_monitor)
            _periodic_metrics_gen.start()
    return _monitor


def get_bg_async() -> BackgroundAsync:
    global _bg_async
    global _bg_async_lock

    with _bg_async_lock:
        if _bg_async is None:
            _bg_async = BackgroundAsync()
    return _bg_async


def get_exchange() -> TopicExchange:
    global _exchange
    global _exchange_lock

    with _exchange_lock:
        if _exchange is None:
            _exchange = TopicExchange()
    return _exchange


def get_trace_handler() -> TraceHandler:
    global _trace_handler
    global _trace_handler_lock

    with _trace_handler_lock:
        if _trace_handler is None:
            _trace_handler = TraceHandler()
    return _trace_handler


def get_renderer() -> Renderer:
    global _renderer
    global _renderer_lock

    with _renderer_lock:
        trace_handler = get_trace_handler()
        if _renderer is None:
            _renderer = AutoRenderer(trace_handler)
    return _renderer


def set_renderer(renderer: Renderer) -> None:
    global _renderer
    global _renderer_lock

    with _renderer_lock:
        _renderer = renderer



from abc import ABC, abstractmethod
from typing import Optional, TypedDict, Union

from ...trace import CaptureOutput


class CaptureVar(TypedDict):
    value: str
    log_prob: Optional[float]


class State(ABC):
    def __init__(self) -> None:
        self.captures: dict[str, Union[CaptureVar, list[CaptureVar]]] = {}
        self.active_role: Optional[str] = None

    @abstractmethod
    def __str__(self) -> str:
        pass

    def apply_capture(
        self, name: str, value: Optional[str], log_prob=Optional[float], is_append: bool = False
    ) -> CaptureOutput:
        if value is None:
            # A "reset" signal
            self.captures.pop(name)
        else:
            var = CaptureVar(value=value, log_prob=log_prob)
            if is_append:
                vars = self.captures.get(name, [])
                if not isinstance(vars, list):
                    vars = [vars]
                vars.append(var)
                self.captures[name] = vars
            else:
                self.captures[name] = var

        return CaptureOutput(
            name=name,
            value=value,
            log_probs=log_prob or float("nan"),
            is_append=is_append,
        )



"""Messages that used between server (usually Jupyter Python kernel) and client.

Messages are required to be added to the model registry for serialization.
"""
from typing import Optional, Dict, Union

from pydantic import BaseModel, Field

from guidance._schema import GenTokenExtra
from ..trace import NodeAttr
import json


_msg_counter: int = -1
class GuidanceMessage(BaseModel):
    """Message sent within Guidance layer."""

    message_id: int = Field(default=None)
    class_name: str = ""

    def __init__(self, **kwargs):
        global _msg_counter

        kwargs["class_name"] = self.__class__.__name__
        if kwargs.get("message_id") is None:
            _msg_counter += 1
            kwargs["message_id"] = _msg_counter
        super().__init__(**kwargs)


class TraceMessage(GuidanceMessage):
    """Update on a trace node."""
    trace_id: int
    parent_trace_id: Optional[int] = None
    node_attr: Optional[NodeAttr] = None


class MetricMessage(GuidanceMessage):
    """Metric that has been emitted."""
    name: str
    value: Union[float, str, list[float], list[str]] = Field(union_mode='left_to_right')
    scalar: bool = True


class ExecutionStartedMessage(GuidanceMessage):
    """Fired when renderer has started trace messages."""


class ExecutionCompletedMessage(GuidanceMessage):
    """Fired when renderer has completed trace messages.

    This functions as the last message sent to client.
    """
    last_trace_id: Optional[int] = None
    is_err: bool = False


class TokensMessage(GuidanceMessage):
    """Fired when trace messages are completed, with tokens for client."""
    trace_id: int
    text: str
    tokens: list[GenTokenExtra]

    def __str__(self):
        return f"message_id={self.message_id} class_name={self.class_name} trace_id={self.trace_id}"


class ResetDisplayMessage(GuidanceMessage):
    """Instructs client to reset the display, removing all output."""
    pass


class ClientReadyMessage(GuidanceMessage):
    """Fired when client is ready to receive messages."""
    pass


class ClientReadyAckMessage(GuidanceMessage):
    """Fired when server acknowledges client readiness."""
    pass


class OutputRequestMessage(GuidanceMessage):
    """Fired when client requests tokens from server."""
    pass


model_registry: Dict[str, type(GuidanceMessage)] = {
    'TraceMessage': TraceMessage,
    'ExecutionStartedMessage': ExecutionStartedMessage,
    'ExecutionCompletedMessage': ExecutionCompletedMessage,
    'ExecutionCompletedOutputMessage': TokensMessage,
    'ResetDisplayMessage': ResetDisplayMessage,
    'ClientReadyMessage': ClientReadyMessage,
    'ClientReadyAckMessage': ClientReadyAckMessage,
    'OutputRequestMessage': OutputRequestMessage,
    'MetricMessage': MetricMessage,
    'TokensMessage': TokensMessage,
}


def serialize_message(message: GuidanceMessage) -> str:
    """ Serializes guidance message.

    Args:
        message: Message to be serialized.

    Returns:
        Serialized message in JSON format.
    """
    message_json = message.model_dump_json(indent=2, serialize_as_any=True)
    return message_json


def deserialize_message(data: str) -> GuidanceMessage:
    """ Deserializes string into a guidance message.

    Args:
        data: JSON string to be deserialized.

    Returns:
        Guidance message.
    """
    data_json = json.loads(data)
    class_name = data_json.get("class_name")
    model_class = model_registry.get(class_name)
    if not model_class:
        raise ValueError(f"Unknown class_name: {class_name}")
    return model_class.model_validate_json(data)


import base64
import wave
from io import BytesIO
from typing import TYPE_CHECKING, Iterator, Literal, Optional, Union

from pydantic import BaseModel, Discriminator, Field, TypeAdapter
from typing_extensions import Annotated, assert_never

from .._ast import (
    ASTNode,
    GenAudio,
    ImageBlob,
    ImageUrl,
    JsonNode,
    LiteralNode,
    RegexNode,
    RoleEnd,
    RoleStart,
    RuleNode,
)
from .._utils import bytes_from
from ..trace import ImageOutput, OutputAttr, TextOutput
from ..trace._trace import AudioOutput
from ._base import Client, Model, State

if TYPE_CHECKING:
    from openai.types.chat import ChatCompletionChunk


def get_role_start(role: str) -> str:
    # ChatML is as good as anything
    return "<|im_start|>" + role + "\n"


def get_role_end(role: str) -> str:
    # ChatML is as good as anything
    return "\n<|im_end|>\n"


class AssistantAudio(BaseModel):
    id: str
    expires_at: int = Field(exclude=True)
    data: str = Field(exclude=True)
    transcript: str = Field(exclude=True)


class AssistantAudioMessage(BaseModel):
    role: Literal["assistant"]
    audio: AssistantAudio


class TextContent(BaseModel):
    type: Literal["text"]
    text: str


class InputAudio(BaseModel):
    data: str
    format: str


class AudioContent(BaseModel):
    type: Literal["input_audio"]
    input_audio: InputAudio


class ImageUrlContentInner(BaseModel):
    url: str


class ImageUrlContent(BaseModel):
    type: Literal["image_url"]
    image_url: ImageUrlContentInner


Content = Annotated[Union[TextContent, AudioContent, ImageUrlContent], Discriminator("type")]


class ContentMessage(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: list[Content]


Message = Union[ContentMessage, AssistantAudioMessage]


class OpenAIState(State):
    def __init__(self) -> None:
        super().__init__()
        self.messages: list[Message] = []
        self.content: list[Content] = []
        self.audio: Optional[AssistantAudio] = None

    def apply_text(self, text: str) -> None:
        if len(self.content) > 0 and isinstance(self.content[-1], TextContent):
            self.content[-1].text += text
        else:
            self.content.append(TextContent(type="text", text=text))

    def get_active_message(state) -> Optional[Message]:
        if state.active_role is None:
            return None
        if state.content and state.audio:
            raise ValueError("Cannot have both content and audio")
        if state.audio:
            return AssistantAudioMessage(
                role=state.active_role,
                audio=state.audio,
            )
        elif state.content:
            return ContentMessage(
                role=state.active_role,
                content=state.content,
            )
        else:
            return None

    def __str__(self) -> str:
        messages = self.messages
        active_message = self.get_active_message()
        if active_message is not None:
            messages = messages + [active_message]
        s = ""
        for i, message in enumerate(messages):
            s += get_role_start(message.role)
            if isinstance(message, AssistantAudioMessage):
                s += "[AUDIO]"
            elif isinstance(message, ContentMessage):
                for content in message.content:
                    if isinstance(content, TextContent):
                        s += content.text
                    elif isinstance(content, ImageUrlContent):
                        s += "[IMAGE]"  # Arbitrary stringification
                    elif isinstance(content, AudioContent):
                        s += "[AUDIO]"  # transcript?
                    else:
                        if TYPE_CHECKING:
                            assert_never(content)
                        raise TypeError(f"Unknown content type: {content}")
            else:
                if TYPE_CHECKING:
                    assert_never(message)
                raise TypeError(f"Unknown message type: {message}")
            if active_message is None or i != len(messages) - 1:
                # For the sake of consistency, don't add role end for the active message
                s += get_role_end(message.role)
        return s


class OpenAIClient(Client[OpenAIState]):
    log_probs: bool = True

    def __init__(
        self,
        model: str,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        try:
            import openai
        except ImportError:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            )
        self.model = model
        self.client = openai.OpenAI(api_key=api_key, **kwargs)

    def run(self, state: OpenAIState, node: ASTNode, **kwargs) -> Iterator[OutputAttr]:
        if not isinstance(node, RoleStart) and state.active_role is None:
            raise ValueError(
                "OpenAI models require an active role (e.g. use `with assistant(): ...`)"
            )
        return super().run(state, node, **kwargs)

    def role_start(self, state: OpenAIState, node: RoleStart, **kwargs) -> Iterator[OutputAttr]:
        state.active_role = node.role
        # TODO: drop this and yield nothing. We need to add this for now as a workaround for the
        # fact that current vis code assumes that there is actually a role start message
        yield TextOutput(value=get_role_start(node.role), is_input=True)

    def role_end(self, state: OpenAIState, node: RoleEnd, **kwargs) -> Iterator[OutputAttr]:
        state.messages.append(state.get_active_message())
        state.audio = None
        state.content = []
        state.active_role = None
        yield from ()

    def text(self, state: OpenAIState, node: LiteralNode, **kwargs) -> Iterator[OutputAttr]:
        state.apply_text(node.value)
        yield TextOutput(value=node.value, is_input=True)

    def rule(self, state: OpenAIState, node: RuleNode, **kwargs) -> Iterator[OutputAttr]:
        if node.stop:
            raise ValueError("Stop condition not yet supported for OpenAI")
        if node.suffix:
            raise ValueError("Suffix not yet supported for OpenAI")
        if node.stop_capture:
            raise ValueError("Save stop text not yet supported for OpenAI")

        kwargs = kwargs.copy()
        if node.temperature:
            kwargs["temperature"] = node.temperature
        if node.max_tokens:
            kwargs["max_tokens"] = node.max_tokens

        chunks = self.run(state, node.value, **kwargs)
        if node.capture:
            buffered_text = ""
            for chunk in chunks:
                # TODO: this isinstance check is pretty darn fragile.
                # ~there must be a better way~
                if isinstance(chunk, TextOutput):
                    buffered_text += chunk.value
                yield chunk
            yield state.apply_capture(
                name=node.capture,
                value=buffered_text,
                log_prob=1,  # TODO
                is_append=node.list_append,
            )
        else:
            yield from chunks

    def regex(self, state: OpenAIState, node: RegexNode, **kwargs) -> Iterator[OutputAttr]:
        if node.regex is not None:
            raise ValueError("Regex not yet supported for OpenAI")
        # We're in unconstrained mode now.
        return self._run(state, **kwargs)

    def json(self, state: OpenAIState, node: JsonNode, **kwargs) -> Iterator[OutputAttr]:
        return self._run(
            state,
            response_format={
                "type": "json_schema",
                "json_schema": {
                    "name": "json_schema",  # TODO?
                    "schema": node.schema,
                    "strict": True,
                },
            },
            **kwargs,
        )

    def _run(self, state: OpenAIState, **kwargs) -> Iterator[OutputAttr]:
        if state.active_role is None:
            # Should never happen?
            raise ValueError(
                "OpenAI models require chat blocks (e.g. use `with assistant(): ...`)"
            )
        if state.active_role != "assistant":
            raise ValueError(
                "OpenAI models can only generate as the assistant (i.e. inside of `with assistant(): ...`)"
            )
        if state.content:
            raise ValueError(
                f"OpenAI models do not support pre-filled assistant messages: got data {state.content}."
            )

        with self.client.chat.completions.create(
            model=self.model,
            messages=TypeAdapter(list[Message]).dump_python(state.messages),  # type: ignore[arg-type]
            logprobs=self.log_probs,
            stream=True,
            **kwargs,
        ) as chunks:
            yield from self._handle_stream(state, chunks)

    def _handle_stream(
        self, state: OpenAIState, chunks: Iterator["ChatCompletionChunk"]
    ) -> Iterator[OutputAttr]:
        audio: Optional[AssistantAudio] = None
        for chunk in chunks:
            choice = chunk.choices[0]
            delta = choice.delta
            if delta.content is not None:
                assert audio is None
                content = delta.content
                if len(content) == 0:
                    continue
                state.apply_text(content)
                if choice.logprobs is not None:
                    # TODO: actually get tokens from this and be less lazy
                    prob = 2.718 ** choice.logprobs.content[0].logprob
                else:
                    prob = float("nan")
                yield TextOutput(value=delta.content, is_generated=True, prob=prob)
            elif getattr(delta, "audio", None) is not None:
                transcript_chunk: Optional[str] = None
                if audio is None:
                    assert delta.audio.get("id") is not None
                    audio = AssistantAudio(
                        id=delta.audio["id"],
                        expires_at=delta.audio.get("expires_at", 0),  # ?
                        transcript=delta.audio.get("transcript", ""),
                        data=delta.audio.get("data", ""),
                    )
                    transcript_chunk = delta.audio.get("transcript")
                else:
                    assert delta.audio.get("id") is None or delta.audio["id"] == audio.id
                    if delta.audio.get("data") is not None:
                        audio.data += delta.audio["data"]
                    if delta.audio.get("transcript") is not None:
                        audio.transcript += delta.audio["transcript"]
                        transcript_chunk = delta.audio["transcript"]
                    if delta.audio.get("expires_at") is not None:
                        assert audio.expires_at == 0
                        audio.expires_at = delta.audio["expires_at"]
                if transcript_chunk is not None:
                    # Why not give the users some transcript? :)
                    yield TextOutput(
                        value=delta.audio["transcript"],
                        is_generated=True,
                    )
            elif delta.function_call is not None:
                raise NotImplementedError("Function calling not yet supported for OpenAI")
            elif delta.tool_calls is not None:
                raise NotImplementedError("Tool calling not yet supported for OpenAI")
            elif delta.refusal is not None:
                raise ValueError(f"OpenAI refused the request: {delta.refusal}")

            if choice.finish_reason is not None:
                break

        if audio is not None:
            assert state.audio is None
            state.audio = audio
            # Create an in-memory WAV file
            wav_buffer = BytesIO()
            with wave.open(wav_buffer, "wb") as wav_file:
                wav_file.setnchannels(1)
                wav_file.setsampwidth(2)  # PCM16 = 2 bytes per sample
                wav_file.setframerate(22050)  # A guess
                wav_file.writeframes(base64.b64decode(audio.data))

            # Get WAV bytes
            wav_bytes = wav_buffer.getvalue()
            yield AudioOutput(value=base64.b64encode(wav_bytes).decode(), is_input=False)


class OpenAIImageClient(OpenAIClient):
    def image_blob(self, state: OpenAIState, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        try:
            import PIL.Image
        except ImportError:
            raise Exception(
                "Please install the Pillow package `pip install Pillow` in order to use images with OpenAI!"
            )

        image_bytes = base64.b64decode(node.data)
        with PIL.Image.open(BytesIO(image_bytes)) as pil_image:
            # Use PIL to infer file format
            # TODO: just store format on ImageOutput type
            format = pil_image.format
            if format is None:
                raise ValueError(f"Cannot upload image with unknown format")

        mime_type = f"image/{format.lower()}"
        state.content.append(
            {"type": "image_url", "image_url": {"url": f"data:{mime_type};base64,{node.data}"}}
        )
        yield ImageOutput(value=node.data, input=True)

    def image_url(self, state: OpenAIState, node: ImageUrl, **kwargs) -> Iterator[OutputAttr]:
        state.content.append({"type": "image_url", "image_url": {"url": node.url}})
        image_bytes = bytes_from(node.url, allow_local=False)
        base64_string = base64.b64encode(image_bytes).decode("utf-8")
        yield ImageOutput(value=base64_string, input=True)


class OpenAIAudioClient(OpenAIClient):
    log_probs: bool = False

    def audio_blob(self, state: OpenAIState, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        format = "wav"  # TODO: infer from node
        state.content.append(
            AudioContent(
                type="input_audio",
                input_audio=InputAudio(
                    data=node.data,
                    format=format,
                ),
            )
        )
        yield AudioOutput(value=node.data, format=format, input=True)

    def gen_audio(self, state: OpenAIState, node: GenAudio, **kwargs) -> Iterator[OutputAttr]:
        yield from self._run(
            state,
            modalities=["text", "audio"],  # Has to be both?
            audio={
                "voice": node.kwargs.get("voice", "alloy"),
                "format": "pcm16",  # Has to be pcm16 for streaming
            },
        )


class OpenAI(Model):
    def __init__(
        self,
        model: str,
        echo: bool = True,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        """Build a new OpenAI model object that represents a model in a given state.

        Parameters
        ----------
        model : str
            The name of the OpenAI model to use (e.g. gpt-4o-mini).
        echo : bool
            If true the final result of creating this model state will be displayed (as HTML in a notebook).
        api_key : None or str
            The OpenAI API key to use for remote requests, passed directly to the `openai.OpenAI` constructor.

        **kwargs :
            All extra keyword arguments are passed directly to the `openai.OpenAI` constructor. Commonly used argument
            names include `base_url` and `organization`
        """

        if "audio-preview" in model:
            client_cls = OpenAIAudioClient
        elif model.startswith("gpt-4o") or model.startswith("o1"):
            client_cls = OpenAIImageClient
        else:
            client_cls = OpenAIClient

        super().__init__(
            client=client_cls(model, api_key=api_key, **kwargs), state=OpenAIState(), echo=echo
        )



import warnings
from json import dumps as json_dumps
from json import loads as json_loads
from typing import Any, Mapping, Optional, Type, Union, cast

import pydantic
from llguidance import JsonCompiler

from .._ast import JsonNode
from .._grammar import capture, token_limit, with_temperature
from ._pydantic import pydantic_to_json_schema

JSONSchema = Union[bool, Mapping[str, Any]]


def json(
    name: Optional[str] = None,
    *,
    schema: Union[
        None,
        str,
        JSONSchema,
        Type[pydantic.BaseModel],
        pydantic.TypeAdapter,
    ] = None,
    temperature: Optional[float] = None,
    max_tokens: Optional[int] = None,
    separators: Optional[tuple[str, str]] = None,
    whitespace_flexible: bool = False,
):
    """Generate valid JSON according to the supplied JSON schema or `pydantic` model.

    Not all parts of `JSON schema <https://json-schema.org/>`_ are supported. Indeed some parts
    (such as bounds on numbers) cannot really be supported in the context of LLM generation.

    Using a JSON schema:

        >>> schema = ''{ "type": "object", "properties": { "a" : {"type": "integer"} } }'
        >>> schema_obj = json.loads(schema)
        >>> lm += json(name="generated_object", schema=schema_obj)
        >>> print(json.loads(lm["generated_object"]))
        { 'a' : 2 }

    Using a ``pydantic.BaseModel``:

        >>> class Schema(BaseModel):
        ...     b: bool
        >>> lm += json(name="generated_object", schema=Schema)
        >>> print(json.loads(lm["generated_object"]))
        { 'b' : False }

    Using a ``pydantic.TypeAdapter``:

        >>> schema = TypeAdapter(list[int])
        >>> lm += json(name="generated_object", schema=schema)
        >>> print(json.loads(lm["generated_object"]))
        [1, 2, 3]

    Parameters
    ----------

    name : str or None
        If this is not None then the the results of the generation will be saved as a variable on
        the Model object (so you can access the result as ``lm["var_name"]``).

    schema : Union[None, Mapping[str, Any], Type[pydantic.BaseModel], pydantic.TypeAdapter]
        One of:
            - None, in which case any valid JSON will be generated
            - A string representing a JSON schema which will be parsed using ``json.loads()``
            - A JSON schema object. This is a JSON schema string which has been passed to ``json.loads()``
            - A subclass of ``pydantic.BaseModel``
            - An instance of ``pydantic.TypeAdapter``
    """
    if schema is False:
        raise ValueError("Unsatisfiable schema: schema is false")
    elif schema is True:
        schema = {}
    elif schema is None:
        # Default schema is empty, "anything goes" schema
        # TODO: consider default being `{"type": "object"}`
        schema = {}
    elif isinstance(schema, pydantic.TypeAdapter) or (
        isinstance(schema, type) and issubclass(schema, pydantic.BaseModel)
    ):
        schema = pydantic_to_json_schema(schema)
    elif isinstance(schema, str):
        from_str = json_loads(schema)
        if not isinstance(from_str, dict):
            raise ValueError("JSON schema string must be a JSON object (i.e. a dictionary)")
        schema = from_str

    if isinstance(schema, Mapping):
        schema = dict(schema)
    else:
        raise TypeError(f"Unsupported schema type: {type(schema)}")

    coerce_one_of = False
    # TODO: decide whether or not to keep this -- it lets us double check that llguidance can handle the schema which isn't necessarily
    # what we want, as llguidance may or may not be the backend we are using. That being said, it's sort of nice to get an exception when
    # you call `json` instead of waiting for generation to fail.
    VALIDATE = True
    if VALIDATE:
        schema_string = json_dumps(schema)
        try:
            compiler = JsonCompiler(
                separators=separators,
                whitespace_flexible=whitespace_flexible,
                coerce_one_of=False,
            )
            compiler.compile(schema_string)
        except ValueError as e:
            if (
                e.args[0]
                == "oneOf constraints are not supported. Enable 'coerce_one_of' option to approximate oneOf with anyOf"
            ):
                warnings.warn(
                    "oneOf not fully supported, falling back to anyOf. This may cause validation errors in some cases."
                )
                compiler = JsonCompiler(
                    separators=separators,
                    whitespace_flexible=whitespace_flexible,
                    coerce_one_of=True,
                )
                compiler.compile(schema_string)
                coerce_one_of = True
            else:
                raise

    if separators is None:
        if whitespace_flexible:
            separators = (",", ":")
        else:
            separators = (", ", ": ")
    item_separator, key_separator = separators

    if schema.get("x-guidance") is None:
        schema["x-guidance"] = {
            "item_separator": item_separator,
            "key_separator": key_separator,
            "whitespace_flexible": whitespace_flexible,
            "coerce_one_of": coerce_one_of,
        }

    node = JsonNode(
        name=name or "json",
        schema=schema,
    )
    if temperature is not None:
        node = with_temperature(node, temperature)
    if max_tokens is not None:
        node = token_limit(node, max_tokens)
    if name is not None:
        node = capture(node, name)
    return node



from typing import Any

from .._base import State


class EngineState(State):
    def __init__(self) -> None:
        super().__init__()
        self.prompt: str = ""
        self.images: list[Any] = []
        self.audio: list[Any] = []
        self.videos: list[Any] = []

    def __str__(self) -> str:
        return self.prompt



import logging
from typing import Optional, Sequence

import numpy as np

from .._schema import EngineOutput, GenToken, GenTokenExtra
from .._utils import softmax
from ..trace import TraceHandler
from ..visual._renderer import DoNothingRenderer
from ._base import Model
from ._engine import Engine, EngineClient, EngineState, Tokenizer

logger = logging.getLogger(__name__)

# TODO: this import pattern happens in a few places, should be cleaned up
try:
    from .. import cpp  # type: ignore[attr-defined]
except ImportError:
    logger.warn("Failed to load guidance.cpp, falling back to Python mirror implementations...")
    from .. import _cpp as cpp


class MockTokenizer(Tokenizer):
    def __init__(self, tokens: Sequence[bytes], special_token_ids: Optional[list[int]] = None):
        super().__init__(tokens, chat_template=None, bos_token_id=0, eos_token_id=0, special_token_ids=special_token_ids)
        self.byte_trie = cpp.ByteTrie(self.tokens, np.arange(len(self.tokens)))

    def encode(self, byte_string: bytes) -> list[int]:
        """Simple greedy tokenizer
        TODO: could be a method on ByteTrie if we want to reuse it
        """
        pos = 0
        tokens = []
        while pos < len(byte_string):
            current_node = self.byte_trie
            last_match = None
            match_pos = pos

            while match_pos < len(byte_string) and current_node.has_child(
                byte_string[match_pos : match_pos + 1]
            ):
                current_node = current_node.child(byte_string[match_pos : match_pos + 1])
                if current_node.value >= 0:
                    last_match = (current_node.value, match_pos + 1)
                match_pos += 1

            if last_match is not None:
                tokens.append(last_match[0])
                pos = last_match[1]
            else:
                raise ValueError(
                    f"Could not find a match for byte {byte_string[pos]} at position {pos}"
                )

        return tokens

    def recode(self, tokens: Sequence[int]) -> list[int]:
        # Make a no-op for now
        return list(tokens)


class MockEngine(Engine):
    def __init__(self, tokenizer, byte_patterns, compute_log_probs, force):
        renderer = DoNothingRenderer(trace_handler=TraceHandler())
        super().__init__(tokenizer, compute_log_probs=compute_log_probs)

        self._valid_mask = np.zeros(len(tokenizer.tokens))
        for i, t in enumerate(tokenizer.tokens):
            try:
                t.decode("utf8")
                self._valid_mask[i] = 1.0
            except:
                pass
        self.force = force
        self.called_temperatures = []

        # allow a single byte pattern to be passed
        if isinstance(byte_patterns, (bytes, str)):
            byte_patterns = [byte_patterns]

        # allow for strings to be passed
        for i, pattern in enumerate(byte_patterns):
            if isinstance(pattern, str):
                byte_patterns[i] = pattern.encode("utf8")

        self.byte_patterns = byte_patterns

        # seed the random number generator
        self._rand_generator = np.random.default_rng(seed=42)

    def sample_with_temperature(self, logits, mask, temperature):
        self.called_temperatures.append(temperature)
        return super().sample_with_temperature(logits, mask, temperature)

    def get_next_token_with_top_k(
        self,
        logits: Optional[np.ndarray],
        logits_lat_ms: Optional[float],
        token_ids: list[int],
        mask: Optional[bytes],
        temperature: float,
        k: int = 1,
        force_return_unmasked_probs: bool = False,
    ) -> EngineOutput:
        self.called_temperatures.append(temperature)
        return super().get_next_token_with_top_k(
            logits, logits_lat_ms, token_ids, mask, temperature, k, force_return_unmasked_probs
        )

    def get_logits(self, token_ids: list[int]) -> np.ndarray:
        """Pretends to compute the logits for the given token state."""
        # build the byte strings
        byte_string = b"".join(self.tokenizer.tokens[i] for i in token_ids)

        # if we are forcing the bytes patterns then don't allow other tokens
        if self.force:
            logits = np.ones(len(self.tokenizer.tokens)) * -np.inf

        # otherwise we randomly generate valid unicode bytes
        else:
            logits = (
                self._rand_generator.standard_normal(len(self.tokenizer.tokens)) * self._valid_mask
            )

        # if we have a pattern that matches then force the next token
        bias = 100.0
        if self.byte_patterns is not None:
            byte_string
            for p in self.byte_patterns:
                if p.startswith(byte_string) and len(p) > len(byte_string):
                    for i in self._get_next_tokens(p[len(byte_string) :]):
                        logits[i] += bias
                    bias /= 2  # if we have multiple matches then they apply with decreasing bias

        return logits

    def get_per_token_topk_probs(
        self, token_ids: list[int], top_k: int = 5
    ) -> list[GenTokenExtra]:
        result_list = []
        if len(token_ids) == 0:
            return result_list

        added_bos = False
        if self.tokenizer.bos_token is not None and token_ids[0] != self.tokenizer.bos_token_id:
            token_ids = [self.tokenizer.bos_token_id] + token_ids
            added_bos = True

        # assume the first token has probability 1.0 because it is the input token
        result_list.append(
            GenTokenExtra(
                token_id=token_ids[0],
                prob=1.0,
                text=self.tokenizer.decode([token_ids[0]]).decode("utf8"),
                top_k=[
                    GenToken(
                        token_id=token_ids[0],
                        prob=1.0,
                        text=self.tokenizer.decode([token_ids[0]]).decode("utf8"),
                    )
                ],
            )
        )

        for i in range(1, len(token_ids)):
            token_id = token_ids[i]
            _logits = self.get_logits(token_ids[:i])
            _probs = softmax(_logits)
            top_k_indices = np.argsort(_logits)[-top_k:][::-1]

            top_k_indices = top_k_indices.tolist()
            if token_ids[i] not in top_k_indices:
                top_k_indices.append(token_id)

            top_k_result = []
            for token_id in top_k_indices:
                top_k_result.append(
                    GenToken(
                        token_id=token_id,
                        prob=_probs[token_id],
                        text=self.tokenizer.decode([token_id]).decode("utf8"),
                    )
                )

            result_list.append(
                GenTokenExtra(
                    token_id=token_id,
                    prob=_probs[token_id],
                    text=self.tokenizer.decode([token_id]).decode("utf-8"),
                    top_k=top_k_result,
                )
            )

        if added_bos:
            result_list = result_list[1:]

        return result_list

    def _get_next_tokens(self, byte_string):
        special_tokens = [
            (self.tokenizer.bos_token_id, self.tokenizer.bos_token),
            (self.tokenizer.eos_token_id, self.tokenizer.eos_token),
        ]
        for i, t in special_tokens:
            # if the byte string starts with a special token then make sure we don't yield any other tokens
            if byte_string.startswith(t):
                yield i
                return
        for i, t in enumerate(self.tokenizer.tokens):
            if byte_string.startswith(t):
                yield i


class Mock(Model):
    def __init__(
        self,
        byte_patterns=[],
        echo=False,
        compute_log_probs=False,
        force=False,
        **kwargs,
    ):
        """Build a new Mock model object that represents a model in a given state."""

        # Our tokens are all bytes and all lowercase letter pairs
        all_lc_pairs = [
            bytes([i, j]) for i in range(ord("a"), ord("z")) for j in range(ord("a"), ord("z"))
        ]
        all_bytes = [bytes([i]) for i in range(256)]
        tokens = [b"<s>"] + all_lc_pairs + all_bytes

        tokenizer = MockTokenizer(tokens, special_token_ids=[0])
        engine = MockEngine(tokenizer, byte_patterns, compute_log_probs, force)

        super().__init__(
            client=EngineClient(engine),
            state=EngineState(),
            echo=echo,
        )


# class MockChat(Mock, Chat):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)



""" Rendering environment detection.

Detection logic is inspired from both plotly and interpretml environment detection.
- https://github.com/plotly/plotly.py
- https://github.com/interpretml/interpret
"""
# TODO(nopdive): Major cloud providers implemented and manually verified.

import os
from pydantic import BaseModel


class EnvFlags(BaseModel):
    """Environment flags - such as if we're in a notebook or cloud."""

    is_notebook: bool = False
    is_cloud: bool = False


class Environment:
    """ Capabilities based environment detection."""

    def __init__(self):
        """ Initializes.

        This will immediately check for which environments are detected.
        """
        self._flags = EnvFlags()
        self._checks = {
            "vscode": _detect_vscode,
            "ipython-zmq": _detect_ipython_zmq,
            "ipython": _detect_ipython,
        }
        envs = []
        for name, update_flags in self._checks.items():
            if update_flags(self._flags):  # in-place operation on flags
                envs.append(name)
        self._detected_envs = envs


    @property
    def detected_envs(self) -> list[str]:
        """ Detected environments (i.e. vscode, ipython-zmq).

        Returns:
            Detected environment names.
        """
        return self._detected_envs


    def is_notebook(self) -> bool:
        """ Determines if the python process is in a notebook.

        Returns:
            True if in notebook.
        """
        return self._flags.is_notebook

    def is_cloud(self) -> bool:
        """ Determines if the python process is in a cloud provider.

        Returns:
            True if in notebook.
        """
        return self._flags.is_cloud

    def is_terminal(self) -> bool:
        """ Determines if the python process not in a notebook (we assume terminal).

        Returns:
            True if in terminal.
        """
        return not self._flags.is_notebook


def _detect_vscode(flags: EnvFlags) -> bool:
    """Detects if called in a vscode process.

    Args:
        flags: Inplace flags to be set.

    Returns:
        True if in vscode environment.
    """
    # NOTE: We don't flag is_notebook since this will be picked up by ipython-zmq here.
    found = "VSCODE_PID" in os.environ
    return found


def _detect_ipython(flags: EnvFlags) -> bool:
    """Detects if called in an IPython environment.
    Mostly derived from stackoverflow below:
    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook

    Args:
        flags: Inplace flags to be set.

    Returns:
        True if in IPython environment.
    """
    found = False
    try:
        from IPython import get_ipython
        found = get_ipython() is not None
    except (NameError, ImportError):  # pragma: no cover
        pass
    return found


def _detect_ipython_zmq(flags: EnvFlags) -> bool:
    """Detects if in an IPython environment using ZMQ (i.e. notebook/qtconsole/lab).

    Mostly derived from stackoverflow below:
    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook/24937408

    Args:
        flags: Inplace flags to be set.

    Returns:
        True if called in IPython notebook or qtconsole.
    """
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            found = True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            found = False  # Terminal running IPython
        else:
            found = False  # Other type (?)
    except (NameError, ImportError):  # pragma: no cover
        found = False  # Probably standard Python interpreter

    flags.is_notebook |= found
    return found


import time
from asyncio import CancelledError
from enum import Enum
from multiprocessing import Manager, Process
from typing import Union, Any
import logging

import psutil

from .._schema import GuidanceEngineMetrics
from ..models import Model

from ..visual import MetricMessage

METRICS_TOPIC = "/metrics"
MISSING_VALUE = 0

logger = logging.getLogger(__name__)


class PeriodicMetricsGenerator:
    def __init__(self, monitor: "Monitor", sleep_sec=0.5):
        self._monitor = monitor
        self._sleep_sec = sleep_sec
        self._task = None
        self._task_cancelled = False
        self._is_paused = False

    def start(self):
        from ..registry import get_bg_async
        bg = get_bg_async()
        self._task = bg.run_async_coroutine(bg.async_task(self._emit())).result()

    def stop(self):
        if self._task is not None:
            self._task.cancel()
            self._task_cancelled = True

    def pause(self):
        """
        Pauses the model by setting the internal _is_paused flag to True.

        This method can be used to temporarily halt the model's operations.
        """
        self._is_paused = True

    def resume(self):
        """
        Resume the model's operation by setting the paused state to False.

        This method changes the internal state of the model to indicate that it is no longer paused.
        """
        self._is_paused = False

    async def _emit(self):
        import asyncio
        import time
        from ..registry import get_exchange

        time_start = time.time()
        while not self._task_cancelled:
            try:
                await asyncio.sleep(self._sleep_sec)

                cpu_percent = self._monitor.get_metric(MonitoringMetric.CPU_USAGE)
                used_ram = self._monitor.get_metric(MonitoringMetric.MEM_USAGE)
                gpu_percent = self._monitor.get_metric(MonitoringMetric.GPU_USAGE)
                gpu_used_vram = self._monitor.get_metric(MonitoringMetric.GPU_USED_MEM)

                if gpu_percent:
                    gpu_percent = max(gpu_percent)
                else:
                    gpu_percent = MISSING_VALUE

                if gpu_used_vram:
                    gpu_used_vram = max(gpu_used_vram)
                else:
                    gpu_used_vram = MISSING_VALUE

                if not cpu_percent:
                    cpu_percent = MISSING_VALUE

                if not used_ram:
                    used_ram = MISSING_VALUE

                time_end = time.time()
                time_elapsed = time_end - time_start

                if not self._is_paused:
                    exchange = get_exchange()
                    exchange.publish(MetricMessage(name="wall time", value=time_elapsed), topic=METRICS_TOPIC)
                    exchange.publish(MetricMessage(name="cpu", value=cpu_percent), topic=METRICS_TOPIC)
                    exchange.publish(MetricMessage(name="ram", value=used_ram), topic=METRICS_TOPIC)
                    exchange.publish(MetricMessage(name="gpu", value=gpu_percent), topic=METRICS_TOPIC)
                    exchange.publish(MetricMessage(name="vram", value=gpu_used_vram), topic=METRICS_TOPIC)
            except CancelledError:
                logger.debug("METRICGEN:canceling")
                break
            except Exception as e:
                logger.debug(f"METRICGEN: {repr(e)}")
                break

        logger.debug("METRICGEN:exiting")


# TODO(nopdive): @hudson-ai needs refactor into engine/class?
class PostExecMetrics:
    def __init__(self, monitor: "Monitor"):
        self._monitor = monitor

    def emit_messages(self, lm: "Model"):
        from ..registry import get_exchange
        exchange = get_exchange()

        token_reduction = self._monitor.get_metric(MonitoringMetric.TOKEN_REDUCTION, lm)

        if token_reduction is not None:
            exchange.publish(MetricMessage(
                name="token reduction",
                value=token_reduction * 100,
            ), topic=METRICS_TOPIC)

        output_tokens = self._monitor.get_metric(MonitoringMetric.OUTPUT_TOKENS, lm)
        if output_tokens is not None:
            exchange.publish(MetricMessage(name="consumed", value=output_tokens), topic=METRICS_TOPIC)

        avg_latency = self._monitor.get_metric(MonitoringMetric.AVG_LATENCY, lm)
        if avg_latency is not None:
            exchange.publish(MetricMessage(name="avg latency", value=avg_latency), topic=METRICS_TOPIC)


class MonitoringMetric(str, Enum):
    CPU_USAGE = "cpu_usage"
    MEM_USAGE = "mem_usage"
    GPU_USAGE = "gpu_usage"
    GPU_USED_MEM = "gpu_used_mem"
    GPU_TOTAL_MEM = "gpu_total_mem"
    INPUT_TOKENS = "input_tokens"
    OUTPUT_TOKENS = "output_tokens"
    BACKTRACK_TOKENS = "backtrack_tokens"
    TOKEN_COUNT = "token_count"
    TOKEN_REDUCTION = "token_reduction"
    AVG_LATENCY = "avg_latency"


ALL_METRICS = [
    MonitoringMetric.CPU_USAGE,
    MonitoringMetric.MEM_USAGE,
    MonitoringMetric.GPU_USAGE,
    MonitoringMetric.GPU_USED_MEM,
    MonitoringMetric.GPU_TOTAL_MEM,
    MonitoringMetric.INPUT_TOKENS,
    MonitoringMetric.OUTPUT_TOKENS,
    MonitoringMetric.BACKTRACK_TOKENS,
    MonitoringMetric.TOKEN_COUNT,
    MonitoringMetric.TOKEN_REDUCTION,
    MonitoringMetric.AVG_LATENCY,
]


def _monitor_fn(
        stop_flag,
        metrics_dict: dict[MonitoringMetric, list],
        max_size: int = 100,
        interval_ms: float = 1000,
):
    # print("Monitoring started")

    to_collect_gpu_stats = False
    has_gpustat = False
    try:
        import gpustat

        has_gpustat = True
    except:
        logger.warning("gpustat is not installed, run `pip install gpustat` to collect GPU stats.")

    if has_gpustat:
        try:
            gpu_stats = gpustat.GPUStatCollection.new_query()
            if len(gpu_stats) > 0:
                # only collect GPU stats if there is at least one GPU
                to_collect_gpu_stats = True
        except:
            logger.warning("Non-Nvidia GPU monitoring is not supported in this version.")

    p = psutil.Process()
    try:
        while not stop_flag.value:
            t0 = time.time()

            cpu_percent = p.cpu_percent() / 100.0
            memory_usage = psutil.virtual_memory()

            metrics_dict[MonitoringMetric.CPU_USAGE].append(cpu_percent)
            metrics_dict[MonitoringMetric.MEM_USAGE].append(memory_usage.used / (1024**3))

            if to_collect_gpu_stats:
                gpu_stats = gpustat.GPUStatCollection.new_query()

                usage = [gpu.utilization / 100.0 for gpu in gpu_stats.gpus]
                mem_usage = [gpu.memory_used for gpu in gpu_stats.gpus]
                mem_total = [gpu.memory_total for gpu in gpu_stats.gpus]

                metrics_dict[MonitoringMetric.GPU_USAGE].append(usage)
                metrics_dict[MonitoringMetric.GPU_USED_MEM].append(mem_usage)
                metrics_dict[MonitoringMetric.GPU_TOTAL_MEM].append(mem_total)

            for metrics in metrics_dict.values():
                if len(metrics) > max_size:
                    metrics.pop(0)

            lat = time.time() - t0

            # sleep for the remaining time of the interval
            sleep_time = interval_ms / 1000.0 - lat
            if sleep_time < 0:
                time.sleep(sleep_time)
    except Exception as e:
        logger.error(f"Error in monitoring: {e}")
        pass

    # print("Monitoring stopped")


class Monitor:
    """Monitoring service to collect necessary metrics for visualization"""

    def __init__(self, engine_metrics: GuidanceEngineMetrics, **kwargs):
        self.engine_metrics = engine_metrics
        self.mp_manager = Manager()

        # use list instead of queue for easily accessing each item, e.g., last item
        self.max_size = kwargs.get("max_size", 100)

        self.metrics_dict = {
            MonitoringMetric.CPU_USAGE: self.mp_manager.list(),
            MonitoringMetric.MEM_USAGE: self.mp_manager.list(),
            MonitoringMetric.GPU_USAGE: self.mp_manager.list(),
            MonitoringMetric.GPU_USED_MEM: self.mp_manager.list(),
            MonitoringMetric.GPU_TOTAL_MEM: self.mp_manager.list(),
        }

        self.stop_flag = self.mp_manager.Value("b", False)
        self.process = None

        self.per_token_metrics = []  # store metrics per token in token list

    def start(self):
        self.process = Process(
            target=_monitor_fn, args=(self.stop_flag, self.metrics_dict, self.max_size)
        )
        self.process.start()
        logger.debug("MONITOR:start")

    def stop(self):
        if self.process:
            self.stop_flag.value = True
            self.process.terminate()

            for metrics in self.metrics_dict.values():
                metrics[:] = []  # NOTE(nopdive): ListProxy does not have .clear method.
        logger.debug("MONITOR:stop")

    def reset(self):
        self.stop()

        for metrics in self.metrics_dict.values():
            metrics.clear()

        self.start()
        logger.debug("MONITOR:reset")

    def get_metrics(
            self, metrics=None, lm: Union["Model", None] = None
    ) -> dict[MonitoringMetric, Any]:
        if metrics is None:
            metrics = ALL_METRICS
        result = {}

        for metric in metrics:
            if metric in [
                MonitoringMetric.CPU_USAGE,
                MonitoringMetric.MEM_USAGE,
                MonitoringMetric.GPU_USAGE,
                MonitoringMetric.GPU_USED_MEM,
                MonitoringMetric.GPU_TOTAL_MEM,
            ]:
                result[metric] = (
                    self.metrics_dict[metric][-1] if len(self.metrics_dict[metric]) > 0 else None
                )
            elif metric == MonitoringMetric.INPUT_TOKENS:
                result[metric] = self.engine_metrics.engine_input_tokens
            elif metric == MonitoringMetric.OUTPUT_TOKENS:
                result[metric] = self.engine_metrics.engine_output_tokens
            elif metric == MonitoringMetric.BACKTRACK_TOKENS:
                result[metric] = self.engine_metrics.engine_backtrack_tokens
            elif metric == MonitoringMetric.TOKEN_COUNT:
                result[metric] = lm.token_count if lm is not None else None
            elif metric == MonitoringMetric.TOKEN_REDUCTION:
                if lm is not None and lm.token_count > 0:
                    result[metric] = 1 - min(1, (lm.metrics.engine_output_tokens / lm.token_count))
                else:
                    result[metric] = None
            elif metric == MonitoringMetric.AVG_LATENCY:
                # TODO(nopdive): Fix this for IR refactor.
                result[metric] = None
        return result

    def get_metric(self, metric: MonitoringMetric, lm: Union["Model", None] = None) -> Any:
        return self.get_metrics([metric], lm)[metric]



from .._ast import GrammarNode, RuleNode
from .._grammar import subgrammar, regex

__all__ = ["subgrammar", "regex", "as_regular_grammar", "lexeme"]

def as_regular_grammar(node: GrammarNode, lexeme=False):
    # TODO: Remove this assertion-only check?
    if isinstance(node, RuleNode):
        rule = node
    else:
        rule = RuleNode("dummy", node)
    assert rule.is_terminal
    return node

def lexeme(body_regex: str, json_string: bool = False):
    if json_string:
        raise NotImplementedError("JSON strings are not supported")
    return regex(body_regex)



from ._base import Model
# from ._engine import Instruct, Chat

# local models
from ._transformers import Transformers, TransformersTokenizer
from ._llama_cpp import LlamaCpp
from ._mock import Mock#, MockChat

# grammarless models (we can't do constrained decoding for them)
# from ._grammarless import Grammarless
# from .vertexai._vertexai import (
#     VertexAI,
#     VertexAIChat,
#     VertexAICompletion,
#     VertexAIInstruct,
# )
# from ._azure_openai import (
#     AzureOpenAI,
# )
# from ._azureai_studio import AzureAIStudioChat
from ._openai import OpenAI
# from ._lite_llm import LiteLLM, LiteLLMChat, LiteLLMInstruct, LiteLLMCompletion
# from ._cohere import Cohere, CohereCompletion, CohereInstruct
# from ._anthropic import Anthropic
# from ._googleai import GoogleAI, GoogleAIChat
# from ._togetherai import (
#     TogetherAI,
#     TogetherAIChat,
#     TogetherAIInstruct,
#     TogetherAICompletion,
# )



""" Metrics that arise from both language models and its execution environment."""

from ._metrics import ALL_METRICS, METRICS_TOPIC
from ._metrics import MonitoringMetric, PostExecMetrics, Monitor, PeriodicMetricsGenerator


import functools
import inspect
import threading
from typing import Any
import weakref
import dataclasses
from contextvars import ContextVar

from ._grammar import string

from ._ast import Function, RuleRefNode, RuleNode
from ._utils import strip_multiline_string_indents, make_weak_bound_method, signature_pop
from .models import Model

_in_stateless_context: ContextVar[bool] = ContextVar("in_stateless_context", default=False)

def guidance(
    f = None,
    *,
    stateless = False,
    cache = False,
    dedent = True,
    model = Model,
):
    """Decorator used to define guidance grammars"""
    # if we are not yet being used as a decorator, then save the args

    if f is None:
        return functools.partial(
            guidance, stateless=stateless, cache=cache, dedent=dedent, model=model,
        )

    # this strips out indentation in multiline strings that aligns with the current python indentation
    if dedent is True or dedent == "python":
        f = strip_multiline_string_indents(f)

    return GuidanceFunction(f, stateless=stateless, cache=cache, model=model)


class GuidanceFunction:
    def __init__(
        self,
        f,
        *,
        stateless = False,
        cache = False,
        model = Model,
    ):
        self.f = f
        self.stateless = stateless
        self.cache = cache
        self.model = model
        self._impl = _decorator(f, stateless=stateless, cache=cache, model=model)
        self._methods: dict[Any, GuidanceMethod] = {}

        # Update self with the wrapped function's metadata
        functools.update_wrapper(self, self._impl)
        # Pretend to be one level of wrapping lower than we are
        self.__wrapped__ = self._impl.__wrapped__

    def __call__(self, *args, **kwargs):
        return self._impl(*args, **kwargs)

    def __get__(self, instance, owner=None, /):
        """
        Return a GuidanceMethod bound to the instance.
        """
        if instance is None:
            return self
        return GuidanceMethod.from_guidance_function(self, instance)

    def __repr__(self):
        return f"<GuidanceFunction {self.__module__}.{self.__qualname__}{self.__signature__}>"

class GuidanceMethod:
    impl_cache = {}
    def __init__(self, impl, instance):
        # Make object that looks like a method (__self__ and __func__) in order to be able to better support weak referencing via weakref.WeakMethod
        # Note we keep a hard reference to the instance to keep it (and therefore our cached impl) alive as long as we are alive
        self.__self__ = instance
        self.__func__ = impl

        # Update self with the wrapped function's metadata
        functools.update_wrapper(self, impl)
        # Pretend to be one level of wrapping lower than we are
        self.__wrapped__ = impl.__wrapped__

    @classmethod
    def from_guidance_function(cls, guidance_function: GuidanceFunction, instance: Any) -> "GuidanceMethod":
        # We can't directly use a weakref.WeakKeyDictionary because those don't really work when the key objects
        # are allowed to change their hash value.

        # Instead use instance hash in addition to identity to make sure we miss the cache if the instance is meaningfully mutated.
        # This should be safe because an id will only be reused after the original object is garbage collected, at which point we
        # should have removed the cache entry (since we use weakref.finalize to remove the cache entry when the instance is deleted).
        key = (guidance_function.f, hash(instance), id(instance))
        try:
            impl = cls.impl_cache[key]
        except KeyError:
            # Make a weak bound method to prevent the instance from being kept alive by the cache
            weak_method = make_weak_bound_method(guidance_function.f, instance)
            impl = _decorator(weak_method, stateless=guidance_function.stateless, cache=guidance_function.cache, model=guidance_function.model)
            cls.impl_cache[key] = impl
            # Clean up the cache when the instance is deleted
            weakref.finalize(instance, cls.impl_cache.pop, key)
        return cls(impl, instance)

    def __call__(self, *args, **kwargs):
        return self.__func__(*args, **kwargs)

    def __repr__(self):
        return f"<bound GuidanceMethod {self.__qualname__} of {self.__self__!r}>"


_null_grammar = string("")


def _decorator(f, *, stateless, cache, model):
    # we cache the function itself if requested
    # do this before updating the wrapper so we can maintain the __wrapped__ chain
    if cache:
        f = functools.cache(f)

    # Use thread local to store the reference to the grammar node for recursive calls
    # Otherwise, shared state between threads may otherwise trick us into thinking we are in a recursive call
    thread_local = threading.local()

    @functools.wraps(f)
    def wrapped(*args, **kwargs):

        # make a stateless grammar if we can
        if stateless is True or (
            callable(stateless) and stateless(*args, **kwargs)
        ):
            # if we have a (deferred) reference set, then we must be in a recursive definition and so we return the reference
            reference = getattr(thread_local, "_self_call_reference_", None)
            if reference is not None:
                return reference

            # otherwise we call the function to generate the grammar
            else:
                # set the stateless context variable so that others can detect that we're currently calling a stateless function
                token = _in_stateless_context.set(True)

                # set a RuleRefNode for recursive calls (only if we don't have arguments that might make caching a bad idea)
                no_args = len(args) + len(kwargs) == 0
                if no_args:
                    thread_local._self_call_reference_ = RuleRefNode()

                try:
                    # call the function to get the grammar node
                    node = f(_null_grammar, *args, **kwargs)
                except:
                    raise
                else:
                    # If we're just wrapping a RuleNode, don't add an extra layer of RuleNode
                    if isinstance(node, RuleNode):
                        rule = dataclasses.replace(node, name=f.__name__)
                    else:
                        rule = RuleNode(name=f.__name__, value=node)
                    # set the reference value with our generated node
                    if no_args:
                        thread_local._self_call_reference_.set_target(rule)
                finally:
                    # Reset the stateless context back to the previous value
                    _in_stateless_context.reset(token)
                    # Clean up the thread local reference
                    if no_args:
                        del thread_local._self_call_reference_

                return rule

        # otherwise must be stateful (which means we can't be inside a select() call)
        else:
            return Function(f, args, kwargs)
 
    # Remove the first argument from the wrapped function since we're going to drop the `lm` argument
    wrapped.__signature__ = signature_pop(inspect.signature(f), 0)

    # attach this as a method of the model class (if given)
    # if model is not None:
    #     setattr(model, f.__name__, f)

    return wrapped



from typing import Union

from .._grammar import repeat
from .._guidance import guidance


@guidance(stateless=True)
def exactly_n_repeats(model, value, n_repeats: int):
    return model + repeat(value, min=n_repeats, max=n_repeats)


@guidance(stateless=True)
def at_most_n_repeats(model, value, n_repeats: int):
    return model + repeat(value, min=0, max=n_repeats)


@guidance(stateless=True)
def sequence(model, value, min_length: int = 0, max_length: Union[int, None] = None):
    # Just an alias for repeat for now -- TODO: remove?
    return model + repeat(value, min=min_length, max=max_length)


@guidance(stateless=True)
def one_or_more(model, value):
    return model + repeat(value, min=1)


@guidance(stateless=True)
def zero_or_more(model, value):
    return model + repeat(value, min=0)



import sys
from typing import (
    Any,
    Callable,
    Literal,
    TypeVar,
    Union,
    overload,
)
if sys.version_info >= (3, 10):
    from typing import ParamSpec, TypeAlias, Concatenate
else:
    from typing_extensions import ParamSpec, TypeAlias, Concatenate

from ._ast import RuleNode, Function
from .models import Model

P = ParamSpec("P")
M: TypeAlias = Any # sort of Union[Model, GrammarNode]?
R = TypeVar("R", bound = Union[Function, RuleNode])
GuidanceWrappable = Callable[Concatenate[M, P], M]
GuidanceFunction = Callable[P, R]
StatefulGuidanceFunction = GuidanceFunction[P, Function]
StatelessGuidanceFunction = GuidanceFunction[P, RuleNode]

@overload
def guidance(
    f: GuidanceWrappable[P],
    *,
    stateless: Literal[False] = False,
    cache: bool = ...,
    dedent: bool = ...,
    model: type[Model] = ...,
) -> StatefulGuidanceFunction[P]:
    ...


@overload
def guidance(
    f: None = None,
    *,
    stateless: Literal[False] = False,
    cache: bool = ...,
    dedent: bool = ...,
    model: type[Model] = ...,
) -> Callable[[GuidanceWrappable[P]], StatefulGuidanceFunction[P]]:
    ...


@overload
def guidance(
    f: GuidanceWrappable[P],
    *,
    stateless: Literal[True],
    cache: bool = ...,
    dedent: bool = ...,
    model: type[Model] = ...,
) -> StatelessGuidanceFunction[P]:
    ...


@overload
def guidance(
    f: None = None,
    *,
    stateless: Literal[True],
    cache: bool = ...,
    dedent: bool = ...,
    model: type[Model] = ...,
) -> Callable[[GuidanceWrappable[P]], StatelessGuidanceFunction[P]]:
    ...


@overload
def guidance(
    f: GuidanceWrappable[P],
    *,
    stateless: Callable[..., bool],
    cache: bool = ...,
    dedent: bool = ...,
    model: type[Model] = ...,
) -> GuidanceFunction[P, Union[Function, RuleNode]]:
    ...


@overload
def guidance(
    f: None = None,
    *,
    stateless: Callable[..., bool],
    cache: bool = ...,
    dedent: bool = ...,
    model: type[Model] = ...,
) -> Callable[[GuidanceWrappable[P]], GuidanceFunction[P, Union[Function, RuleNode]]]:
    ...



import pathlib
import typing
import base64

from .._guidance import guidance
from .._utils import bytes_from
from .._ast import AudioBlob, GenAudio


@guidance
def audio(lm, src: typing.Union[str, pathlib.Path, bytes], allow_local: bool = True):
    bytes_data = bytes_from(src, allow_local=allow_local)
    base64_string = base64.b64encode(bytes_data).decode('utf-8')
    lm += AudioBlob(data=base64_string)
    return lm


@guidance
def gen_audio(lm, **kwargs):
    return lm + GenAudio(kwargs=kwargs)



import ast
import asyncio
import json
import inspect
import sys
import textwrap
import types
import weakref
import functools
import numpy as np
import logging
from typing import Union, cast
import pathlib
import urllib
import http
import re

logger = logging.getLogger(__name__)


def bytes_from(src: Union[str, pathlib.Path, bytes], allow_local: bool) -> bytes:
    if isinstance(src, str) and re.match(r"[^:/]+://", src):
        with urllib.request.urlopen(src) as response:
            response = cast(http.client.HTTPResponse, response)
            bytes_data = response.read()

    # ...from a local path
    elif allow_local and (isinstance(src, str) or isinstance(src, pathlib.Path)):
        with open(src, "rb") as f:
            bytes_data = f.read()

    # ...from audio file bytes
    elif isinstance(src, bytes):
        bytes_data = src

    else:
        raise Exception(f"Unable to load bytes from {src}!")

    return bytes_data


class _Rewrite(ast.NodeTransformer):
    def __init__(self, source_lines):
        self.source_lines = source_lines
        self.indentation = [None for _ in source_lines]

    def visit_JoinedStr(self, node):
        for value in node.values:
            if isinstance(value, ast.Constant) and isinstance(value.value, str):
                self._dedent_constant(value, node.lineno)
        return node

    def visit_Constant(self, node):
        if isinstance(node.value, str) and "\n" in node.value:
            self._dedent_constant(node, node.lineno)
        return node

    def _dedent_constant(self, node, lineno):
        start_lineno = lineno - 1
        start_line = self.source_lines[start_lineno]
        indent = len(start_line) - len(start_line.lstrip())

        if indent > 0:
            new_lines = []
            for line in node.value.split("\n"):
                if line.startswith(" " * indent):
                    new_lines.append(line[indent:])
                else:
                    new_lines.append(line)
            node.value = "\n".join(new_lines)

class normalize_notebook_stdout_stderr:
    """Remaps stdout and stderr back to their normal selves from what ipykernel did to them.

    Based on: https://github.com/ipython/ipykernel/issues/795
    """

    def __enter__(self):
        normal_stdout = sys.__stdout__.fileno()
        self.restore_stdout = None
        if (
            getattr(sys.stdout, "_original_stdstream_copy", normal_stdout)
            != normal_stdout
        ):
            self.restore_stdout = sys.stdout._original_stdstream_copy
            sys.stdout._original_stdstream_copy = normal_stdout

        normal_stderr = sys.__stderr__.fileno()
        self.restore_stderr = None
        if (
            getattr(sys.stderr, "_original_stdstream_copy", normal_stderr)
            != normal_stderr
        ):
            self.restore_stderr = sys.stderr._original_stdstream_copy
            sys.stderr._original_stdstream_copy = normal_stderr

    def __exit__(self, exc_type, exc_value, traceback):
        if self.restore_stdout is not None:
            sys.stderr._original_stdstream_copy = self.restore_stdout
        if self.restore_stderr is not None:
            sys.stderr._original_stdstream_copy = self.restore_stderr


def strip_multiline_string_indents(f):
    source = textwrap.dedent(inspect.getsource(f))
    blanks = "\n" * f.__code__.co_firstlineno  # pad the source so the lines in the file line up for the debugger
    source = blanks + "\n".join(source.splitlines()[1:])  # remove the decorator first line.

    # define the external closure variables so f.__closure__ will match our recompiled version
    if len(f.__code__.co_freevars) > 0:
        raise Exception(
            "You currently must use @guidance(dedent=False) for closure functions (function nested within other functions that reference the outer functions variables)!"
        )
#       lines = source.split("\n")
#       lines[0] = "def __outer__closure_wrap():"
#       lines[1] = (
#           "    "
#           + ",".join(f.__code__.co_freevars)
#           + " = "
#           + ",".join("None" for _ in f.__code__.co_freevars)
#       )
#       source = "    \n".join(
#           lines
#       )  # TODO: this does not quite work because new_code_obj is now the __outer__closure_wrap() function...could be fixed with work...

    old_code_obj = f.__code__
    old_ast = ast.parse(source)
    r = _Rewrite(source.split("\n"))
    new_ast = r.visit(old_ast)
    new_code_obj = compile(new_ast, old_code_obj.co_filename, "exec")

    # find the code block
    for i in range(len(new_code_obj.co_consts)):
        if str(type(new_code_obj.co_consts[i])) == "<class 'code'>":
            break

    # create a new function based on the modified code
    new_f = types.FunctionType(
        new_code_obj.co_consts[i],
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    new_f.__kwdefaults__ = f.__kwdefaults__
    new_f.__qualname__ = f.__qualname__
    new_f.__annotations__ = f.__annotations__
    new_f.__doc__ = f.__doc__
    new_f.__module__ = f.__module__
    return new_f

def make_weak_bound_method(f, instance):
    instance_ref = weakref.ref(instance)
    instance_repr = repr(instance)
    @functools.wraps(f) # ish
    def weak_bound_f(*args, **kwargs):
        instance = instance_ref()
        if instance is None:
            raise ReferenceError(f"Lost reference to {instance_repr} and cannot bind {f} to it.")
        method = types.MethodType(f, instance)
        return method(*args, **kwargs)

    # remove the first argument from the wrapped function since it is now bound
    weak_bound_f.__signature__ = signature_pop(inspect.signature(f), 0)
    return weak_bound_f

def signature_pop(signature, index):
    params = list(signature.parameters.values())
    params.pop(index)
    return signature.replace(parameters=params)

class JupyterComm:
    def __init__(
        self, target_id, ipython_handle, callback=None, on_open=None, mode="register"
    ):
        from ipykernel.comm import Comm

        self.target_name = "guidance_interface_target_" + target_id
        # print("TARGET NAME", self.target_name)
        self.callback = callback
        self.jcomm = None
        self.ipython_handle = ipython_handle
        self.addd = 1
        self.send_queue = asyncio.Queue()
        self.open_event = asyncio.Event()
        self.is_open = False
        asyncio.get_event_loop().create_task(self._send_loop())
        if mode == "register":
            # log("REGISTERING", self.target_name)
            # asyncio.get_event_loop().create_task(self._register())
            def comm_opened(comm, open_msg):
                # log("OPENED")
                self.addd = 2
                self.jcomm = comm
                self.is_open = True
                self.jcomm.on_msg(self._fire_callback)
                self.open_event.set()
                self._fire_callback({"content": {"data": {"event": "opened"}}})

            self.ipython_handle.kernel.comm_manager.register_target(
                self.target_name, comm_opened
            )
            # get_ipython().kernel.comm_manager.register_target(self.target_name, comm_opened) # noqa: F821
        elif mode == "open":
            # log("OPENING", self.target_name)
            self.jcomm = Comm(target_name=self.target_name)
            self.jcomm.on_msg(self._fire_callback)
            # self._fire_callback({"content": {"data": "opened"}})
        else:
            raise Exception("Passed mode must be either 'open' or 'register'!")

    def clear_send_queue(self):
        while not self.send_queue.empty():
            self.send_queue.get_nowait()
            self.send_queue.task_done()

    def _fire_callback(self, msg):
        self.callback(msg["content"]["data"])

    def send(self, data):
        self.send_queue.put_nowait(data)

    async def _send_loop(self):
        while True:
            # log("SENDING_LOOP")
            if self.jcomm is None:
                self.open_event.clear()
                await self.open_event.wait()
            data = await self.send_queue.get()
            # log("SENDING_LOOP got one!")
            self.jcomm.send({"data": json.dumps(data)})

    # async def _waiting_send(self, data):
    #     #log("SENDING", self.jcomm, data)

    #     # await the open event if needed
    #     if self.jcomm is None:
    #         self.open_event.clear()
    #         await self.open_event.wait()
    #     #log("SENDING_now", self.jcomm, data)
    #     self.jcomm.send({"data": json.dumps(data)}) # we encode the JSON so iPython doesn't mess it up


# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def is_interactive():
    import __main__ as main

    return not hasattr(main, "__file__")


def log_softmax(array: np.ndarray, axis: int = -1) -> np.ndarray:
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.log_softmax.html
    array_maxs: np.ndarray = np.amax(array, axis=axis, keepdims=True)
    if array_maxs.ndim > 0:
        array_maxs[~np.isfinite(array_maxs)] = 0
    elif not np.isfinite(array_maxs):
        array_maxs = np.zeros(array_maxs.shape)
    subtract_maxs = array - array_maxs
    exp = np.exp(subtract_maxs)
    # suppress warnings about log of zero
    with np.errstate(divide="ignore"):
        summed = np.sum(exp, axis=axis, keepdims=True)
        out = np.log(summed)
    return subtract_maxs - out


def softmax(array: np.ndarray, axis: int = -1) -> np.ndarray:
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.softmax.html
    array_maxs = np.amax(array, axis=axis, keepdims=True)
    exp_x_shifted = np.exp(array - array_maxs)
    return exp_x_shifted / np.sum(exp_x_shifted, axis=axis, keepdims=True)


def pydantic_no_default_repr(obj, target_fields=None):
    if target_fields is None:
        records = (
            f'{getattr(obj, name)!r}'
            for name, field in obj.model_fields.items()
            if getattr(obj, name) != field.default and not field.exclude
        )
    else:
        records = (
            f'{getattr(obj, name)!r}'
            for name, field in obj.model_fields.items()
            if getattr(obj, name) != field.default and not field.exclude and name in target_fields
        )
    out = f'{type(obj).__name__}:{":".join(records)}'
    return out


def pydantic_no_default_str(obj, target_fields=None):
    if target_fields is None:
        records = (
            f'{getattr(obj, name)!s}'
            for name, field in obj.model_fields.items()
            if getattr(obj, name) != field.default and not field.exclude
        )
    else:
        records = (
            f'{getattr(obj, name)!s}'
            for name, field in obj.model_fields.items()
            if getattr(obj, name) != field.default and not field.exclude and name in target_fields
        )
    out = "\n".join(records)
    return out


def log_init(s: str):
    logger.debug(f"INIT:{s}")
    pass


def log_copy(s: str):
    logger.debug(f"COPY:{s}")
    pass


def log_cleanup(s: str):
    logger.debug(f"CLEANUP:{s}")
    pass

def to_utf8_or_bytes_string(_bytes: bytes) -> str:
    """
    Converts a byte sequence to a UTF-8 string if possible. If the byte sequence
    cannot be decoded as UTF-8, it returns the string representation of the byte sequence.

    Parameters
    ----------
    _bytes : bytes
        The byte sequence to be converted.

    Returns
    -------
    str
        The decoded UTF-8 string or the string representation of the byte sequence
        if UTF-8 decoding fails.
    """
    try:
        return _bytes.decode("utf-8")
    except UnicodeDecodeError:
        return str(_bytes)


from base64 import b64decode
from io import BytesIO
from typing import Iterator

from ..._ast import GrammarNode, ImageBlob, LiteralNode, RoleEnd, RoleStart
from ...trace import ImageOutput, OutputAttr, TextOutput
from .._base import Client
from ._engine import Engine
from ._state import EngineState


class EngineClient(Client[EngineState]):
    def __init__(self, engine: Engine):
        self.engine = engine
        self.chat_template = self.engine.get_chat_template()

    def get_role_start(self, role: str) -> str:
        if self.chat_template is None:
            raise ValueError("Cannot use roles without a chat template")
        return self.chat_template.get_role_start(role)

    def get_role_end(self, role: str) -> str:
        if self.chat_template is None:
            raise ValueError("Cannot use roles without a chat template")
        return self.chat_template.get_role_end(role)

    def role_start(self, state: EngineState, node: RoleStart, **kwargs) -> Iterator[OutputAttr]:
        state.active_role = node.role
        # TODO: mark these as special tokens..?
        yield from self.run(state, LiteralNode(value=self.get_role_start(node.role)), **kwargs)

    def role_end(self, state: EngineState, node: RoleEnd, **kwargs) -> Iterator[OutputAttr]:
        state.active_role = None
        # TODO: mark these as special tokens..?
        yield from self.run(state, LiteralNode(value=self.get_role_end(node.role)), **kwargs)

    def text(self, state: EngineState, node: LiteralNode, **kwargs) -> Iterator[OutputAttr]:
        state.prompt += node.value
        yield TextOutput(value=node.value, is_input=True)

    def grammar(self, state: EngineState, node: GrammarNode, **kwargs) -> Iterator[OutputAttr]:
        engine_gen = self.engine(
            state,
            node.ll_grammar(),
            ensure_bos_token=True,
            echo=False,
        )

        delayed_bytes = b""
        for chunk in engine_gen:
            new_bytes = chunk.new_bytes
            new_text, delayed_bytes = partial_decode(new_bytes)

            # Update the state
            state.prompt += new_text
            yield TextOutput(value=new_text, token_count=chunk.new_token_count, is_generated=True)

            # TODO -- rewrite engine internals to make sure chunk.{generated,fast_forwarded}_tokens aren't empty...
            # # TODO: GenTokenExtra
            # for token in chunk.generated_tokens:
            #     yield TextOutput(
            #         value=token.text,  # TODO: this should really be the token bytes
            #         is_generated=True,
            #         token_count=1,
            #         prob=token.prob,
            #         tokens=[token],  # TODO: drop this
            #     )
            # for token in chunk.force_forwarded_tokens:
            #     yield TextOutput(
            #         value=token.text,  # TODO: this should really be the token bytes
            #         is_generated=False,
            #         token_count=1,
            #         prob=token.prob,
            #         tokens=[token],  # TODO: drop this
            #     )

            # # TODO: yield some kind of backtrack signal?

            for name in chunk.capture_groups.keys():
                values = chunk.capture_groups[name]
                log_probs = chunk.capture_group_log_probs[name]
                if isinstance(values, list):
                    assert isinstance(log_probs, list)
                    assert len(values) == len(log_probs)
                    for value, log_prob in zip(values, log_probs):
                        yield state.apply_capture(
                            name, value.decode("utf-8"), log_prob=log_prob, is_append=True
                        )
                else:
                    assert isinstance(log_probs, float)
                    yield state.apply_capture(
                        name, values.decode("utf-8"), log_prob=log_probs, is_append=False
                    )

        if delayed_bytes:
            raise RuntimeError("Shouldn't have any delayed bytes left...")


class Llama3VisionClient(EngineClient):
    def image_blob(self, state: EngineState, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        try:
            import PIL.Image
        except ImportError:
            raise Exception(
                "Please install the Pillow package `pip install Pillow` in order to use images with Llama3!"
            )

        image_bytes = b64decode(node.data)
        pil_image = PIL.Image.open(BytesIO(image_bytes))
        state.images.append(pil_image)
        state.prompt += "<|image|>"

        yield ImageOutput(value=node.data, input=True)


class Phi3VisionClient(EngineClient):
    def image_blob(self, state: EngineState, node: ImageBlob, **kwargs) -> Iterator[OutputAttr]:
        try:
            import PIL.Image
        except ImportError:
            raise Exception(
                "Please install the Pillow package `pip install Pillow` in order to use images with Llama3!"
            )

        image_bytes = b64decode(node.data)
        pil_image = PIL.Image.open(BytesIO(image_bytes))

        if pil_image in state.images:
            ix = state.images.index(pil_image) + 1
        else:
            state.images.append(pil_image)
            ix = len(state.images)
        state.prompt += f"<|image_{ix}|>"

        yield ImageOutput(value=node.data, input=True)


def partial_decode(data: bytes) -> tuple[str, bytes]:
    try:
        return (data.decode("utf-8"), b"")
    except UnicodeDecodeError as e:
        valid_part = data[: e.start].decode("utf-8")
        delayed_part = data[e.start :]
    return (valid_part, delayed_part)



"""UI and other visual UX considerations.

Users should have few reasons to be accessing this module.
"""

from ._message import GuidanceMessage, TraceMessage, ResetDisplayMessage, ClientReadyMessage, ClientReadyAckMessage
from ._message import ExecutionCompletedMessage, TokensMessage, MetricMessage, OutputRequestMessage
from ._message import ExecutionStartedMessage
from ._renderer import AutoRenderer, JupyterWidgetRenderer, Renderer
from ._message import serialize_message, deserialize_message
from ._trace import trace_node_to_str, display_trace_tree, trace_node_to_html
from ._exchange import TopicExchange